@@ -6,9 +6,9 @@
 
 #include "IndexMap.h"
 #include <algorithm>
+#include <dolfinx/common/sort.h>
 #include <functional>
 #include <numeric>
-#include <unordered_map>
 
 using namespace dolfinx;
 using namespace dolfinx::common;
@@ -16,29 +16,10 @@ using namespace dolfinx::common;
 namespace
 {
 //-----------------------------------------------------------------------------
-
-/// Compute the owner on the neighbourgood communicator of ghost indices
-std::vector<int>
-compute_ghost_owners(const std::vector<int>& ghost_pos_recv_fwd,
-                     const std::vector<int>& displs_recv_fwd)
-{
-  std::vector<int> owners;
-  std::transform(ghost_pos_recv_fwd.cbegin(), ghost_pos_recv_fwd.cend(),
-                 std::back_inserter(owners),
-                 [&displs_recv_fwd](auto ghost_pos)
-                 {
-                   auto it
-                       = std::upper_bound(displs_recv_fwd.cbegin(),
-                                          displs_recv_fwd.cend(), ghost_pos);
-                   return std::distance(displs_recv_fwd.cbegin(), it) - 1;
-                 });
-  return owners;
-}
-//----------------------------------------------------------------------------
-
 /// Compute the owning rank of ghost indices
-std::vector<int> get_ghost_ranks(MPI_Comm comm, std::int32_t local_size,
-                                 const xtl::span<const std::int64_t>& ghosts)
+[[maybe_unused]] std::vector<int>
+get_ghost_ranks(MPI_Comm comm, std::int32_t local_size,
+                const xtl::span<const std::int64_t>& ghosts)
 {
   int mpi_size = -1;
   MPI_Comm_size(comm, &mpi_size);
@@ -107,6 +88,8 @@ compute_owned_shared(MPI_Comm comm, const xtl::span<const std::int64_t>& ghosts,
   // Send number of my 'ghost indices' to each owner, and receive number
   // of my 'owned indices' that are ghosted on other ranks
   std::vector<int> in_edges_num(src_ranks.size());
+  in_edges_num.reserve(1);
+  out_edges_num.reserve(1);
   MPI_Neighbor_alltoall(out_edges_num.data(), 1, MPI_INT, in_edges_num.data(),
                         1, MPI_INT, comm);
 
@@ -130,13 +113,6 @@ compute_owned_shared(MPI_Comm comm, const xtl::span<const std::int64_t>& ghosts,
     }
   }
 
-  // Work-around for OpenMPI bug. Some version of OpenMPI crash if
-  // in/out_edges_num array is empty.
-  if (in_edges_num.empty())
-    in_edges_num.resize(1);
-  if (out_edges_num.empty())
-    out_edges_num.resize(1);
-
   // May have repeated shared indices with different processes
   std::vector<std::int64_t> recv_indices(recv_disp.back());
   MPI_Neighbor_alltoallv(send_indices.data(), out_edges_num.data(),
@@ -152,6 +128,114 @@ compute_owned_shared(MPI_Comm comm, const xtl::span<const std::int64_t>& ghosts,
 //-----------------------------------------------------------------------------
 } // namespace
 
+//-----------------------------------------------------------------------------
+std::vector<int32_t> dolfinx::common::compute_owned_indices(
+    const xtl::span<const std::int32_t>& indices, const IndexMap& map)
+{
+  // Split indices into those owned by this process and those that are
+  // ghosts. `ghost_indices` contains the position of the ghost in
+  // map.ghosts()
+  std::vector<std::int32_t> owned;
+  std::vector<std::int32_t> ghost_indices;
+  const int size_local = map.size_local();
+
+  // Get number of owned and ghost indices in indicies list to reserve
+  // vectors
+  const int num_owned = std::count_if(indices.begin(), indices.end(),
+                                      [size_local](std::int32_t index)
+                                      { return index < size_local; });
+  const int num_ghost = indices.size() - num_owned;
+  owned.reserve(num_owned);
+  ghost_indices.reserve(num_ghost);
+  std::for_each(indices.begin(), indices.end(),
+                [&owned, &ghost_indices, size_local](const std::int32_t index)
+                {
+                  if (index < size_local)
+                    owned.push_back(index);
+                  else
+                    ghost_indices.push_back(index - size_local);
+                });
+
+  // Create an AdjacencyList whose nodes are the processes in the
+  // neighborhood and the links for a given process are the ghosts
+  // (global numbering) in `indices` owned by that process.
+  MPI_Comm reverse_comm = map.comm(IndexMap::Direction::reverse);
+  std::vector<std::int32_t> dest_ranks
+      = dolfinx::MPI::neighbors(reverse_comm)[1];
+
+  std::vector<int> ghost_owner_rank;
+  {
+    std::vector<int> neighbors = dolfinx::MPI::neighbors(
+        map.comm(common::IndexMap::Direction::forward))[0];
+    ghost_owner_rank = map.ghost_owners();
+    std::transform(ghost_owner_rank.cbegin(), ghost_owner_rank.cend(),
+                   ghost_owner_rank.begin(),
+                   [&neighbors](auto r) { return neighbors[r]; });
+  }
+
+  const std::vector<std::int64_t>& ghosts = map.ghosts();
+  std::vector<std::int64_t> ghosts_to_send;
+  std::vector<std::int32_t> ghosts_per_proc(dest_ranks.size(), 0);
+
+  // Loop through all destination ranks in the neighborhood
+  for (std::size_t dest_rank_index = 0; dest_rank_index < dest_ranks.size();
+       ++dest_rank_index)
+  {
+    // Loop through all ghost indices on this rank
+    for (std::int32_t ghost_index : ghost_indices)
+    {
+      // Check if the ghost is owned by the destination rank. If so, add
+      // that ghost so it is sent to the correct process.
+      if (ghost_owner_rank[ghost_index] == dest_ranks[dest_rank_index])
+      {
+        ghosts_to_send.push_back(ghosts[ghost_index]);
+        ghosts_per_proc[dest_rank_index]++;
+      }
+    }
+  }
+  // Create a list of partial sums of the number of ghosts per process
+  // and create the AdjacencyList
+  std::vector<int> send_disp(dest_ranks.size() + 1, 0);
+  std::partial_sum(ghosts_per_proc.begin(), ghosts_per_proc.end(),
+                   std::next(send_disp.begin(), 1));
+  const graph::AdjacencyList<std::int64_t> data_out(std::move(ghosts_to_send),
+                                                    std::move(send_disp));
+
+  // Communicate ghosts on this process in `indices` back to their owners
+  const graph::AdjacencyList<std::int64_t> data_in
+      = dolfinx::MPI::neighbor_all_to_all(reverse_comm, data_out);
+
+  // Get the local index from the global indices received from other
+  // processes and add to `owned`
+  const std::vector<std::int64_t>& global_indices = map.global_indices();
+  std::vector<std::pair<std::int64_t, std::int32_t>> global_to_local;
+  global_to_local.reserve(global_indices.size());
+  for (auto idx : global_indices)
+  {
+    global_to_local.push_back(
+        {idx, static_cast<std::int32_t>(global_to_local.size())});
+  }
+  std::sort(global_to_local.begin(), global_to_local.end());
+  std::transform(
+      data_in.array().cbegin(), data_in.array().cend(),
+      std::back_inserter(owned),
+      [&global_to_local](std::int64_t global_index)
+      {
+        auto it = std::lower_bound(
+            global_to_local.begin(), global_to_local.end(),
+            typename decltype(global_to_local)::value_type(global_index, 0),
+            [](auto& a, auto& b) { return a.first < b.first; });
+        assert(it != global_to_local.end() and it->first == global_index);
+        return it->second;
+      });
+
+  // Sort `owned` and remove non-unique entries (we could have received
+  // the same ghost from multiple other processes)
+  dolfinx::radix_sort(xtl::span(owned));
+  owned.erase(std::unique(owned.begin(), owned.end()), owned.end());
+
+  return owned;
+}
 //-----------------------------------------------------------------------------
 std::tuple<std::int64_t, std::vector<std::int32_t>,
            std::vector<std::vector<std::int64_t>>,
@@ -215,10 +299,9 @@ common::stack_index_maps(
   // Create neighborhood communicator
   MPI_Comm comm;
   MPI_Dist_graph_create_adjacent(
-      maps.at(0).first.get().comm(IndexMap::Direction::forward),
-      in_neighbors.size(), in_neighbors.data(), MPI_UNWEIGHTED,
-      out_neighbors.size(), out_neighbors.data(), MPI_UNWEIGHTED, MPI_INFO_NULL,
-      false, &comm);
+      maps.at(0).first.get().comm(), in_neighbors.size(), in_neighbors.data(),
+      MPI_UNWEIGHTED, out_neighbors.size(), out_neighbors.data(),
+      MPI_UNWEIGHTED, MPI_INFO_NULL, false, &comm);
 
   int indegree(-1), outdegree(-2), weighted(-1);
   MPI_Dist_graph_neighbors_count(comm, &indegree, &outdegree, &weighted);
@@ -226,6 +309,7 @@ common::stack_index_maps(
   // Figure out how much data to receive from each neighbor
   const int num_my_rows = indices.size();
   std::vector<int> num_rows_recv(indegree);
+  num_rows_recv.reserve(1);
   MPI_Neighbor_allgather(&num_my_rows, 1, MPI_INT, num_rows_recv.data(), 1,
                          MPI_INT, comm);
 
@@ -256,8 +340,17 @@ common::stack_index_maps(
   {
     const int bs = maps[f].second;
     const std::vector<std::int64_t>& ghosts = maps[f].first.get().ghosts();
-    const std::vector<int>& ghost_owners
-        = maps[f].first.get().ghost_owner_rank();
+
+    std::vector<int> ghost_owners;
+    {
+      std::vector<int> neighbors = dolfinx::MPI::neighbors(
+          maps[f].first.get().comm(common::IndexMap::Direction::forward))[0];
+      ghost_owners = maps[f].first.get().ghost_owners();
+      std::transform(ghost_owners.cbegin(), ghost_owners.cend(),
+                     ghost_owners.begin(),
+                     [&neighbors](auto r) { return neighbors[r]; });
+    }
+
     for (std::size_t i = 0; i < ghosts.size(); ++i)
     {
       for (int j = 0; j < bs; ++j)
@@ -276,7 +369,8 @@ common::stack_index_maps(
 //-----------------------------------------------------------------------------
 //-----------------------------------------------------------------------------
 IndexMap::IndexMap(MPI_Comm comm, std::int32_t local_size)
-    : _comm_owner_to_ghost(MPI_COMM_NULL), _comm_ghost_to_owner(MPI_COMM_NULL)
+    : _comm(comm), _comm_owner_to_ghost(MPI_COMM_NULL),
+      _comm_ghost_to_owner(MPI_COMM_NULL), _displs_recv_fwd(1, 0)
 {
   // Get global offset (index), using partial exclusive reduction
   std::int64_t offset = 0;
@@ -300,41 +394,44 @@ IndexMap::IndexMap(MPI_Comm comm, std::int32_t local_size)
   // Create communicators with empty neighborhoods
   MPI_Comm comm0, comm1;
   std::vector<int> ranks(0);
+  // NOTE: create uniform weights as a workaround to issue
+  // https://github.com/pmodels/mpich/issues/5764
+  std::vector<int> weights(0);
   MPI_Dist_graph_create_adjacent(comm, ranks.size(), ranks.data(),
-                                 MPI_UNWEIGHTED, ranks.size(), ranks.data(),
-                                 MPI_UNWEIGHTED, MPI_INFO_NULL, false, &comm0);
+                                 weights.data(), ranks.size(), ranks.data(),
+                                 weights.data(), MPI_INFO_NULL, false, &comm0);
   MPI_Dist_graph_create_adjacent(comm, ranks.size(), ranks.data(),
-                                 MPI_UNWEIGHTED, ranks.size(), ranks.data(),
-                                 MPI_UNWEIGHTED, MPI_INFO_NULL, false, &comm1);
+                                 weights.data(), ranks.size(), ranks.data(),
+                                 weights.data(), MPI_INFO_NULL, false, &comm1);
   _comm_owner_to_ghost = dolfinx::MPI::Comm(comm0, false);
   _comm_ghost_to_owner = dolfinx::MPI::Comm(comm1, false);
   _shared_indices = std::make_unique<graph::AdjacencyList<std::int32_t>>(0);
 }
 //-----------------------------------------------------------------------------
-IndexMap::IndexMap(MPI_Comm mpi_comm, std::int32_t local_size,
+IndexMap::IndexMap(MPI_Comm comm, std::int32_t local_size,
                    const xtl::span<const int>& dest_ranks,
                    const xtl::span<const std::int64_t>& ghosts,
                    const xtl::span<const int>& src_ranks)
-    : _comm_owner_to_ghost(MPI_COMM_NULL), _comm_ghost_to_owner(MPI_COMM_NULL),
-      _ghosts(ghosts.begin(), ghosts.end())
+    : _comm(comm), _comm_owner_to_ghost(MPI_COMM_NULL),
+      _comm_ghost_to_owner(MPI_COMM_NULL), _ghosts(ghosts.begin(), ghosts.end())
 {
   assert(size_t(ghosts.size()) == src_ranks.size());
   assert(std::equal(src_ranks.begin(), src_ranks.end(),
-                    get_ghost_ranks(mpi_comm, local_size, _ghosts).begin()));
+                    get_ghost_ranks(comm, local_size, _ghosts).begin()));
 
   // Get global offset (index), using partial exclusive reduction
   std::int64_t offset = 0;
   const std::int64_t local_size_tmp = (std::int64_t)local_size;
   MPI_Request request_scan;
-  MPI_Iexscan(&local_size_tmp, &offset, 1, MPI_INT64_T, MPI_SUM, mpi_comm,
+  MPI_Iexscan(&local_size_tmp, &offset, 1, MPI_INT64_T, MPI_SUM, comm,
               &request_scan);
 
   // Send local size to sum reduction to get global size
   MPI_Request request;
-  MPI_Iallreduce(&local_size_tmp, &_size_global, 1, MPI_INT64_T, MPI_SUM,
-                 mpi_comm, &request);
+  MPI_Iallreduce(&local_size_tmp, &_size_global, 1, MPI_INT64_T, MPI_SUM, comm,
+                 &request);
 
-  // Build vector of src ranks for ghosts, i.e. the ranks that own the
+  // Build set of src ranks for ghosts, i.e. the ranks that own the
   // callers ghosts
   std::vector<std::int32_t> halo_src_ranks(src_ranks.begin(), src_ranks.end());
   std::sort(halo_src_ranks.begin(), halo_src_ranks.end());
@@ -345,37 +442,37 @@ IndexMap::IndexMap(MPI_Comm mpi_comm, std::int32_t local_size,
   // Create communicators with directed edges: (0) owner -> ghost,
   // (1) ghost -> owner
   {
-    // Dummy weight to work around a bug in older MPICH versions. Issue
-    // appears in MPICH v3.3.2 (Ubuntu 20.04) and not in v3.4 (Ubuntu
-    // 20.10). Would prefer to pass MPI_UNWEIGHTED.
-    std::vector<int> weights(std::max(halo_src_ranks.size(), dest_ranks.size()),
-                             1);
+    // NOTE: create uniform weights as a workaround to issue
+    // https://github.com/pmodels/mpich/issues/5764
+    std::vector<int> src_weights(halo_src_ranks.size(), 1);
+    std::vector<int> dest_weights(dest_ranks.size(), 1);
 
     MPI_Comm comm0;
     MPI_Dist_graph_create_adjacent(
-        mpi_comm, halo_src_ranks.size(), halo_src_ranks.data(), weights.data(),
-        dest_ranks.size(), dest_ranks.data(), weights.data(), MPI_INFO_NULL,
-        false, &comm0);
+        comm, halo_src_ranks.size(), halo_src_ranks.data(), src_weights.data(),
+        dest_ranks.size(), dest_ranks.data(), dest_weights.data(),
+        MPI_INFO_NULL, false, &comm0);
     _comm_owner_to_ghost = dolfinx::MPI::Comm(comm0, false);
 
     MPI_Comm comm1;
-    MPI_Dist_graph_create_adjacent(
-        mpi_comm, dest_ranks.size(), dest_ranks.data(), weights.data(),
-        halo_src_ranks.size(), halo_src_ranks.data(), weights.data(),
-        MPI_INFO_NULL, false, &comm1);
+    MPI_Dist_graph_create_adjacent(comm, dest_ranks.size(), dest_ranks.data(),
+                                   dest_weights.data(), halo_src_ranks.size(),
+                                   halo_src_ranks.data(), src_weights.data(),
+                                   MPI_INFO_NULL, false, &comm1);
     _comm_ghost_to_owner = dolfinx::MPI::Comm(comm1, false);
   }
 
   // Map ghost owner rank to the rank on neighborhood communicator
   int myrank = -1;
-  MPI_Comm_rank(mpi_comm, &myrank);
+  MPI_Comm_rank(comm, &myrank);
+  assert(std::find(src_ranks.begin(), src_ranks.end(), myrank)
+         == src_ranks.end());
   std::vector<std::int32_t> ghost_owners(ghosts.size());
   std::transform(src_ranks.cbegin(), src_ranks.cend(), ghost_owners.begin(),
-                 [&halo_src_ranks, myrank](auto src)
+                 [&halo_src_ranks](auto src)
                  {
                    // Get rank of owner on the neighborhood communicator
                    // (rank of out edge on _comm_owner_to_ghost)
-                   assert(src != myrank);
                    auto it = std::find(halo_src_ranks.cbegin(),
                                        halo_src_ranks.cend(), src);
                    assert(it != halo_src_ranks.end());
@@ -430,18 +527,11 @@ IndexMap::IndexMap(MPI_Comm mpi_comm, std::int32_t local_size,
 
   // Build array that maps ghost indicies to a position in the recv
   // (forward scatter) and send (reverse scatter) buffers
-  std::vector<std::int32_t> tmp_displs = _displs_recv_fwd;
   _ghost_pos_recv_fwd.resize(ghost_owners.size());
-  for (std::size_t i = 0; i < ghost_owners.size(); ++i)
-    _ghost_pos_recv_fwd[i] = tmp_displs[ghost_owners[i]]++;
-
-  // Add '1' for OpenMPI bug
-  if (_sizes_recv_fwd.empty())
-    _sizes_recv_fwd.push_back(0);
-
-  // Add '1' for OpenMPI bug
-  if (_sizes_send_fwd.empty())
-    _sizes_send_fwd.push_back(0);
+  std::transform(ghost_owners.cbegin(), ghost_owners.cend(),
+                 _ghost_pos_recv_fwd.begin(),
+                 [displs = _displs_recv_fwd](auto owner) mutable
+                 { return displs[owner]++; });
 }
 //-----------------------------------------------------------------------------
 std::array<std::int64_t, 2> IndexMap::local_range() const noexcept
@@ -534,27 +624,22 @@ IndexMap::scatter_fwd_ghost_positions() const noexcept
   return _ghost_pos_recv_fwd;
 }
 //-----------------------------------------------------------------------------
-std::vector<int> IndexMap::ghost_owner_rank() const
+std::vector<int> IndexMap::ghost_owners() const
 {
-  int indegree(-1), outdegree(-2), weighted(-1);
-  MPI_Dist_graph_neighbors_count(_comm_owner_to_ghost.comm(), &indegree,
-                                 &outdegree, &weighted);
-  std::vector<int> neighbors_in(indegree), neighbors_out(outdegree);
-  MPI_Dist_graph_neighbors(_comm_owner_to_ghost.comm(), indegree,
-                           neighbors_in.data(), MPI_UNWEIGHTED, outdegree,
-                           neighbors_out.data(), MPI_UNWEIGHTED);
-
-  // Compute index owner on neighbourhood comm
-  const std::vector<int> ghost_owners
-      = compute_ghost_owners(_ghost_pos_recv_fwd, _displs_recv_fwd);
-  std::vector<std::int32_t> owners(ghost_owners.size());
-  std::transform(ghost_owners.cbegin(), ghost_owners.cend(), owners.begin(),
-                 [&neighbors_in](auto ghost_owner)
-                 { return neighbors_in[ghost_owner]; });
+  std::vector<int> owners(_ghost_pos_recv_fwd.size());
+  std::transform(
+      _ghost_pos_recv_fwd.begin(), _ghost_pos_recv_fwd.end(), owners.begin(),
+      [&displs = _displs_recv_fwd](auto pos)
+      {
+        auto it = std::upper_bound(displs.begin(), displs.end(), pos);
+        return std::distance(displs.begin(), it) - 1;
+      });
 
   return owners;
 }
 //----------------------------------------------------------------------------
+MPI_Comm IndexMap::comm() const { return _comm.comm(); }
+//----------------------------------------------------------------------------
 MPI_Comm IndexMap::comm(Direction dir) const
 {
   switch (dir)
@@ -595,7 +680,7 @@ std::map<std::int32_t, std::set<int>> IndexMap::compute_shared_indices() const
   // other ranks which also ghost the index.
 
   std::vector<std::int64_t> fwd_sharing_data;
-  std::vector<int> fwd_sharing_offsets{0};
+  std::vector<int> fwd_sharing_offsets = {0};
   for (std::int32_t p = 0; p < _shared_indices->num_nodes(); ++p)
   {
     for (std::int32_t idx : _shared_indices->links(p))
@@ -622,10 +707,11 @@ std::map<std::int32_t, std::set<int>> IndexMap::compute_shared_indices() const
   // Send sharing rank data from owner to ghosts
 
   // Send data size to send, and get to-receive sizes
-  std::vector<int> send_sizes(outdegree, 0);
-  std::vector<int> recv_sizes(indegree);
+  std::vector<int> send_sizes(outdegree, 0), recv_sizes(indegree);
   std::adjacent_difference(fwd_sharing_offsets.begin() + 1,
                            fwd_sharing_offsets.end(), send_sizes.begin());
+  send_sizes.reserve(1);
+  recv_sizes.reserve(1);
   MPI_Neighbor_alltoall(send_sizes.data(), 1, MPI_INT, recv_sizes.data(), 1,
                         MPI_INT, _comm_owner_to_ghost.comm());
 
@@ -635,12 +721,6 @@ std::map<std::int32_t, std::set<int>> IndexMap::compute_shared_indices() const
                    recv_offsets.begin() + 1);
   std::vector<std::int64_t> recv_data(recv_offsets.back());
 
-  // Work-around for OpenMPI
-  if (send_sizes.empty())
-    send_sizes.resize(1);
-  if (recv_sizes.empty())
-    recv_sizes.resize(1);
-
   // Start data exchange
   MPI_Request request;
   MPI_Ineighbor_alltoallv(
@@ -650,15 +730,29 @@ std::map<std::int32_t, std::set<int>> IndexMap::compute_shared_indices() const
 
   // For my ghosts, add owning rank to list of sharing ranks
   const std::int32_t size_local = this->size_local();
-  const std::vector<int> ghost_owners = this->ghost_owner_rank();
+
+  std::vector<int> ghost_owners;
+  {
+    std::vector<int> neighbors
+        = dolfinx::MPI::neighbors(_comm_owner_to_ghost.comm())[0];
+    ghost_owners = this->ghost_owners();
+    std::transform(ghost_owners.cbegin(), ghost_owners.cend(),
+                   ghost_owners.begin(),
+                   [&neighbors](auto r) { return neighbors[r]; });
+  }
+
   for (std::size_t i = 0; i < _ghosts.size(); ++i)
     shared_indices[size_local + i].insert(ghost_owners[i]);
 
   // Build map from global index to local index for ghosts
-  std::unordered_map<std::int64_t, std::int32_t> ghosts;
+  std::vector<std::pair<std::int64_t, std::int32_t>> ghosts;
   ghosts.reserve(_ghosts.size());
-  for (std::size_t i = 0; i < _ghosts.size(); ++i)
-    ghosts.emplace(_ghosts[i], i + size_local);
+  for (auto idx : _ghosts)
+  {
+    ghosts.push_back(
+        {idx, static_cast<std::int32_t>(size_local + ghosts.size())});
+  }
+  std::sort(ghosts.begin(), ghosts.end());
 
   // Wait for all-to-all to complete
   MPI_Wait(&request, MPI_STATUS_IGNORE);
@@ -668,18 +762,215 @@ std::map<std::int32_t, std::set<int>> IndexMap::compute_shared_indices() const
   MPI_Comm_rank(_comm_owner_to_ghost.comm(), &myrank);
   for (std::size_t i = 0; i < recv_data.size();)
   {
-    auto it = ghosts.find(recv_data[i]);
-    assert(it != ghosts.end());
+    auto it = std::lower_bound(
+        ghosts.begin(), ghosts.end(),
+        typename decltype(ghosts)::value_type(recv_data[i], 0),
+        [](auto& a, auto& b) { return a.first < b.first; });
+    assert(it != ghosts.end() and it->first == recv_data[i]);
+
     const std::int32_t idx = it->second;
     const int set_size = recv_data[i + 1];
-    for (int j = 0; j < set_size; j++)
+    for (int j = 0; j < set_size; ++j)
     {
       if (recv_data[i + 2 + j] != myrank)
-        shared_indices[idx].insert(recv_data[i + 2 + +j]);
+        shared_indices[idx].insert(recv_data[i + 2 + j]);
     }
     i += set_size + 2;
   }
 
   return shared_indices;
 }
-//-----------------------------------------------------------------------------
\ No newline at end of file
+//-----------------------------------------------------------------------------
+std::pair<IndexMap, std::vector<std::int32_t>>
+IndexMap::create_submap(const xtl::span<const std::int32_t>& indices) const
+{
+  if (!indices.empty() and indices.back() >= this->size_local())
+  {
+    throw std::runtime_error(
+        "Unowned index detected when creating sub-IndexMap");
+  }
+
+  MPI_Comm comm = this->comm(Direction::forward);
+
+  int myrank = 0;
+  MPI_Comm_rank(comm, &myrank);
+
+  // --- Step 1: Compute new offest for this rank and new global size
+
+  std::int64_t local_size = indices.size();
+  std::int64_t offset = 0;
+  MPI_Request request_offset;
+  MPI_Iexscan(&local_size, &offset, 1, MPI_INT64_T, MPI_SUM, comm,
+              &request_offset);
+
+  std::int64_t size_global = 0;
+  MPI_Request request_size;
+  MPI_Iallreduce(&local_size, &size_global, 1, MPI_INT64_T, MPI_SUM, comm,
+                 &request_size);
+
+  // --- Step 2: Create array from old to new index for owned indices,
+  // setting entries to -1 if they are not in the new map
+
+  std::vector<std::int32_t> old_to_new_index(this->size_local(), -1);
+  for (std::size_t i = 0; i < indices.size(); ++i)
+    old_to_new_index[indices[i]] = i;
+
+  // --- Step 3: Compute the destination that the new index map will
+  // send to (fwd scatter) and build the new shared_indices adjacency
+  // list
+
+  // Loop over ranks that ghost data in the original map
+  std::vector<int> ranks_old_to_new_send(_shared_indices->num_nodes(), -1);
+  std::vector<std::int32_t> shared_indices_data, shared_indices_off(1, 0);
+  for (std::int32_t r_old = 0; r_old < _shared_indices->num_nodes(); ++r_old)
+  {
+    // For indices sent to old rank r_old, add the new index
+    int num_indices = 0;
+    for (std::int32_t idx_old : _shared_indices->links(r_old))
+    {
+      if (auto idx = old_to_new_index[idx_old]; idx >= 0)
+      {
+        shared_indices_data.push_back(idx);
+        ++num_indices;
+      }
+    }
+
+    // If indices in the new map will be sent to r_old, update
+    // old-to-new map and update offset array
+    if (num_indices > 0)
+    {
+      ranks_old_to_new_send[r_old] = shared_indices_off.size() - 1;
+      shared_indices_off.push_back(shared_indices_off.back() + num_indices);
+    }
+  }
+
+  // Create new shared_indices adjacency list
+  shared_indices_data.shrink_to_fit();
+  shared_indices_off.shrink_to_fit();
+  auto shared_indices = std::make_unique<graph::AdjacencyList<std::int32_t>>(
+      std::move(shared_indices_data), std::move(shared_indices_off));
+
+  // --- Step 4: Create sizes_send_fwd array
+
+  MPI_Wait(&request_offset, MPI_STATUS_IGNORE);
+
+  // TODO: Can we avoid this step and pack the buffer directly?
+  // Build array of new global indices for indices in the new map
+  std::vector<std::int64_t> global_indices_new(this->size_local(), -1);
+  for (std::size_t i = 0; i < indices.size(); ++i)
+    global_indices_new[indices[i]] = i + offset;
+
+  // --- Step 5: Send new global indices to ranks that ghost indices on
+  // this rank
+
+  const std::vector<std::int32_t>& send_indices
+      = this->scatter_fwd_indices().array();
+  std::vector<std::int64_t> buffer_send_fwd(send_indices.size());
+  std::transform(
+      send_indices.cbegin(), send_indices.cend(), buffer_send_fwd.begin(),
+      [&global_indices_new](auto idx) { return global_indices_new[idx]; });
+
+  MPI_Request request1;
+  std::vector<std::int64_t> buffer_recv_fwd(this->num_ghosts());
+  MPI_Datatype data_type = MPI_INT64_T;
+  this->scatter_fwd_begin(xtl::span<const std::int64_t>(buffer_send_fwd),
+                          data_type, request1,
+                          xtl::span<std::int64_t>(buffer_recv_fwd));
+  this->scatter_fwd_end(request1);
+
+  // --- Step 6: Determine which ranks that send data to this rank for
+  // *this will also send data to the new map on this rank
+
+  // Count number of ghost from each rank
+  std::vector<std::int32_t> ranks_old_to_new_recv(_sizes_recv_fwd.size(), -1),
+      displs_recv_fwd(1, 0), sizes_recv_fwd_new(_sizes_recv_fwd.size(), 0);
+  for (std::size_t r_old = 0; r_old < _sizes_recv_fwd.size(); ++r_old)
+  {
+    // Count number of ghosts owned by rank r_old
+    assert(r_old + 1 < _displs_recv_fwd.size());
+    sizes_recv_fwd_new[r_old] = std::count_if(
+        std::next(buffer_recv_fwd.cbegin(), _displs_recv_fwd[r_old]),
+        std::next(buffer_recv_fwd.cbegin(), _displs_recv_fwd[r_old + 1]),
+        [](auto x) { return x >= 0; });
+
+    if (sizes_recv_fwd_new[r_old] > 0)
+    {
+      // Will receive data from r_old
+      ranks_old_to_new_recv[r_old] = displs_recv_fwd.size() - 1;
+      displs_recv_fwd.push_back(displs_recv_fwd.back()
+                                + sizes_recv_fwd_new[r_old]);
+    }
+  }
+
+  // --- Step 7: Build ghost_pos_recv_fwd for the new map
+
+  // Build list of ghosts in the new map, and compute the new owning rank
+  std::vector<std::int64_t> ghosts;
+  std::vector<std::int32_t> ghost_owner, new_to_old_ghost;
+  for (std::size_t i = 0; i < _ghosts.size(); ++i)
+  {
+    // Get new index for old _ghosts[i]
+    std::size_t buffer_pos = _ghost_pos_recv_fwd[i];
+    if (auto index = buffer_recv_fwd[buffer_pos]; index >= 0)
+    {
+      // If index is in the new map, add index and compute owning rank
+      // (rank index on new map)
+      ghosts.push_back(index);
+      new_to_old_ghost.push_back(i);
+      auto it = std::upper_bound(_displs_recv_fwd.cbegin(),
+                                 _displs_recv_fwd.cend(), buffer_pos);
+      const int r_old = std::distance(_displs_recv_fwd.cbegin(), it) - 1;
+      ghost_owner.push_back(ranks_old_to_new_recv[r_old]);
+    }
+  }
+
+  // Compute ghost_pos_recv_fwd
+  std::vector<std::int32_t> ghost_pos_recv_fwd(ghost_owner.size());
+  std::transform(ghost_owner.cbegin(), ghost_owner.cend(),
+                 ghost_pos_recv_fwd.begin(),
+                 [displs = displs_recv_fwd](auto owner) mutable
+                 { return displs[owner]++; });
+
+  // Step 8: Create neighbourhood communicators for the new map
+
+  // Get global src/dest from forward comm (owner-to-ghost) from
+  // original map
+  const auto [src_ranks, dest_ranks] = dolfinx::MPI::neighbors(comm);
+  std::vector<int> in_ranks, out_ranks;
+  for (std::size_t r = 0; r < ranks_old_to_new_recv.size(); ++r)
+    if (ranks_old_to_new_recv[r] >= 0)
+      in_ranks.push_back(src_ranks[r]);
+  for (std::size_t r = 0; r < ranks_old_to_new_send.size(); ++r)
+    if (ranks_old_to_new_send[r] >= 0)
+      out_ranks.push_back(dest_ranks[r]);
+
+  // NOTE: create uniform weights as a workaround to issue
+  // https://github.com/pmodels/mpich/issues/5764
+  std::vector<int> in_weights(in_ranks.size(), 1);
+  std::vector<int> out_weights(out_ranks.size(), 1);
+
+  MPI_Comm comm0, comm1;
+  MPI_Dist_graph_create_adjacent(comm, in_ranks.size(), in_ranks.data(),
+                                 in_weights.data(), out_ranks.size(),
+                                 out_ranks.data(), out_weights.data(),
+                                 MPI_INFO_NULL, false, &comm0);
+  MPI_Dist_graph_create_adjacent(comm, out_ranks.size(), out_ranks.data(),
+                                 out_weights.data(), in_ranks.size(),
+                                 in_ranks.data(), in_weights.data(),
+                                 MPI_INFO_NULL, false, &comm1);
+
+  // Wait for the MPI_Iallreduce to complete
+  MPI_Wait(&request_size, MPI_STATUS_IGNORE);
+
+  displs_recv_fwd.shrink_to_fit();
+  ghosts.shrink_to_fit();
+  new_to_old_ghost.shrink_to_fit();
+
+  return {IndexMap({offset, offset + local_size}, size_global, _comm.comm(),
+                   dolfinx::MPI::Comm(comm0, false),
+                   dolfinx::MPI::Comm(comm1, false), std::move(displs_recv_fwd),
+                   std::move(ghost_pos_recv_fwd), std::move(ghosts),
+                   std::move(shared_indices)),
+          std::move(new_to_old_ghost)};
+}
+//-----------------------------------------------------------------------------
