@@ -8,8 +8,9 @@
 
 #include "DofMap.h"
 #include "Form.h"
+#include "FunctionSpace.h"
 #include "utils.h"
-#include <dolfinx/fem/FunctionSpace.h>
+#include <dolfinx/common/utils.h>
 #include <dolfinx/graph/AdjacencyList.h>
 #include <dolfinx/la/utils.h>
 #include <dolfinx/mesh/Geometry.h>
@@ -27,42 +28,43 @@ namespace dolfinx::fem::impl
 /// local indices. Rows (bc0) and columns (bc1) with Dirichlet
 /// conditions are zeroed. Markers (bc0 and bc1) can be empty if not bcs
 /// are applied. Matrix is not finalised.
-
-template <typename T>
+template <typename T, typename U>
 void assemble_matrix(
-    const std::function<int(std::int32_t, const std::int32_t*, std::int32_t,
-                            const std::int32_t*, const T*)>& mat_set_values,
-    const Form<T>& a, const xtl::span<const T>& constants,
-    const array2d<T>& coeffs, const std::vector<bool>& bc0,
-    const std::vector<bool>& bc1);
+    U mat_set_values, const Form<T>& a, const xtl::span<const T>& constants,
+    const std::map<std::pair<IntegralType, int>,
+                   std::pair<xtl::span<const T>, int>>& coefficients,
+    const xtl::span<const std::int8_t>& bc0,
+    const xtl::span<const std::int8_t>& bc1);
 
 /// Execute kernel over cells and accumulate result in matrix
-template <typename T>
+template <typename T, typename U>
 void assemble_cells(
-    const std::function<int(std::int32_t, const std::int32_t*, std::int32_t,
-                            const std::int32_t*, const T*)>& mat_set,
-    const mesh::Geometry& geometry,
-    const xtl::span<const std::int32_t>& active_cells,
+    U mat_set, const mesh::Geometry& geometry,
+    const xtl::span<const std::int32_t>& cells,
     const std::function<void(const xtl::span<T>&,
                              const xtl::span<const std::uint32_t>&,
                              std::int32_t, int)>& dof_transform,
-    const graph::AdjacencyList<std::int32_t>& dofmap0, const int bs0,
+    const graph::AdjacencyList<std::int32_t>& dofmap0, int bs0,
     const std::function<void(const xtl::span<T>&,
                              const xtl::span<const std::uint32_t>&,
                              std::int32_t, int)>& dof_transform_to_transpose,
-    const graph::AdjacencyList<std::int32_t>& dofmap1, const int bs1,
-    const std::vector<bool>& bc0, const std::vector<bool>& bc1,
+    const graph::AdjacencyList<std::int32_t>& dofmap1, int bs1,
+    const xtl::span<const std::int8_t>& bc0,
+    const xtl::span<const std::int8_t>& bc1,
     const std::function<void(T*, const T*, const T*, const double*, const int*,
                              const std::uint8_t*)>& kernel,
-    const array2d<T>& coeffs, const xtl::span<const T>& constants,
+    const xtl::span<const T>& coeffs, int cstride,
+    const xtl::span<const T>& constants,
     const xtl::span<const std::uint32_t>& cell_info)
 {
+  if (cells.empty())
+    return;
+
   // Prepare cell geometry
   const graph::AdjacencyList<std::int32_t>& x_dofmap = geometry.dofmap();
+  const std::size_t num_dofs_g = geometry.cmap().dim();
 
-  // FIXME: Add proper interface for num coordinate dofs
-  const std::size_t num_dofs_g = x_dofmap.num_links(0);
-  const xt::xtensor<double, 2>& x_g = geometry.x();
+  xtl::span<const double> x_g = geometry.x();
 
   // Iterate over active cells
   const int num_dofs0 = dofmap0.links(0).size();
@@ -73,19 +75,22 @@ void assemble_cells(
   const xtl::span<T> _Ae(Ae);
   std::vector<double> coordinate_dofs(3 * num_dofs_g);
 
-  for (std::int32_t c : active_cells)
+  // Iterate over active cells
+  for (std::size_t index = 0; index < cells.size(); ++index)
   {
+    std::int32_t c = cells[index];
+
     // Get cell coordinates/geometry
     auto x_dofs = x_dofmap.links(c);
     for (std::size_t i = 0; i < x_dofs.size(); ++i)
     {
-      std::copy_n(xt::row(x_g, x_dofs[i]).begin(), 3,
-                  std::next(coordinate_dofs.begin(), 3 * i));
+      common::impl::copy_N<3>(std::next(x_g.begin(), 3 * x_dofs[i]),
+                              std::next(coordinate_dofs.begin(), 3 * i));
     }
 
     // Tabulate tensor
     std::fill(Ae.begin(), Ae.end(), 0);
-    kernel(Ae.data(), coeffs.row(c).data(), constants.data(),
+    kernel(Ae.data(), coeffs.data() + index * cstride, constants.data(),
            coordinate_dofs.data(), nullptr, nullptr);
 
     dof_transform(_Ae, cell_info, c, ndim1);
@@ -127,16 +132,15 @@ void assemble_cells(
       }
     }
 
-    mat_set(dofs0.size(), dofs0.data(), dofs1.size(), dofs1.data(), Ae.data());
+    mat_set(dofs0, dofs1, Ae);
   }
 }
 
 /// Execute kernel over exterior facets and  accumulate result in Mat
-template <typename T>
+template <typename T, typename U>
 void assemble_exterior_facets(
-    const std::function<int(std::int32_t, const std::int32_t*, std::int32_t,
-                            const std::int32_t*, const T*)>& mat_set,
-    const mesh::Mesh& mesh, const xtl::span<const std::int32_t>& active_facets,
+    U mat_set, const mesh::Mesh& mesh,
+    const xtl::span<const std::pair<std::int32_t, int>>& facets,
     const std::function<void(const xtl::span<T>&,
                              const xtl::span<const std::uint32_t>&,
                              std::int32_t, int)>& dof_transform,
@@ -145,21 +149,21 @@ void assemble_exterior_facets(
                              const xtl::span<const std::uint32_t>&,
                              std::int32_t, int)>& dof_transform_to_transpose,
     const graph::AdjacencyList<std::int32_t>& dofmap1, int bs1,
-    const std::vector<bool>& bc0, const std::vector<bool>& bc1,
+    const xtl::span<const std::int8_t>& bc0,
+    const xtl::span<const std::int8_t>& bc1,
     const std::function<void(T*, const T*, const T*, const double*, const int*,
                              const std::uint8_t*)>& kernel,
-    const array2d<T>& coeffs, const xtl::span<const T>& constants,
-    const xtl::span<const std::uint32_t>& cell_info,
-    const std::function<std::uint8_t(std::size_t)>& get_perm)
+    const xtl::span<const T>& coeffs, int cstride,
+    const xtl::span<const T>& constants,
+    const xtl::span<const std::uint32_t>& cell_info)
 {
-  const int tdim = mesh.topology().dim();
+  if (facets.empty())
+    return;
 
   // Prepare cell geometry
   const graph::AdjacencyList<std::int32_t>& x_dofmap = mesh.geometry().dofmap();
-
-  // FIXME: Add proper interface for num coordinate dofs
-  const std::size_t num_dofs_g = x_dofmap.num_links(0);
-  const xt::xtensor<double, 2>& x_g = mesh.geometry().x();
+  const std::size_t num_dofs_g = mesh.geometry().cmap().dim();
+  xtl::span<const double> x_g = mesh.geometry().x();
 
   // Data structures used in assembly
   std::vector<double> coordinate_dofs(3 * num_dofs_g);
@@ -170,43 +174,30 @@ void assemble_exterior_facets(
   std::vector<T> Ae(ndim0 * ndim1);
   const xtl::span<T> _Ae(Ae);
 
-  // Iterate over all facets
-  auto f_to_c = mesh.topology().connectivity(tdim - 1, tdim);
-  assert(f_to_c);
-  auto c_to_f = mesh.topology().connectivity(tdim, tdim - 1);
-  assert(c_to_f);
-
-  for (std::int32_t f : active_facets)
+  for (std::size_t index = 0; index < facets.size(); ++index)
   {
-    auto cells = f_to_c->links(f);
-    assert(cells.size() == 1);
-
-    // Get local index of facet with respect to the cell
-    auto facets = c_to_f->links(cells[0]);
-    auto it = std::find(facets.begin(), facets.end(), f);
-    assert(it != facets.end());
-    const int local_facet = std::distance(facets.begin(), it);
+    std::int32_t cell = facets[index].first;
+    int local_facet = facets[index].second;
 
     // Get cell coordinates/geometry
-    auto x_dofs = x_dofmap.links(cells[0]);
+    auto x_dofs = x_dofmap.links(cell);
     for (std::size_t i = 0; i < x_dofs.size(); ++i)
     {
-      std::copy_n(xt::row(x_g, x_dofs[i]).begin(), 3,
-                  std::next(coordinate_dofs.begin(), 3 * i));
+      common::impl::copy_N<3>(std::next(x_g.begin(), 3 * x_dofs[i]),
+                              std::next(coordinate_dofs.begin(), 3 * i));
     }
 
     // Tabulate tensor
-    std::uint8_t perm = get_perm(cells[0] * facets.size() + local_facet);
     std::fill(Ae.begin(), Ae.end(), 0);
-    kernel(Ae.data(), coeffs.row(cells[0]).data(), constants.data(),
-           coordinate_dofs.data(), &local_facet, &perm);
+    kernel(Ae.data(), coeffs.data() + index * cstride, constants.data(),
+           coordinate_dofs.data(), &local_facet, nullptr);
 
-    dof_transform(_Ae, cell_info, cells[0], ndim1);
-    dof_transform_to_transpose(_Ae, cell_info, cells[0], ndim0);
+    dof_transform(_Ae, cell_info, cell, ndim1);
+    dof_transform_to_transpose(_Ae, cell_info, cell, ndim0);
 
     // Zero rows/columns for essential bcs
-    auto dofs0 = dofmap0.links(cells[0]);
-    auto dofs1 = dofmap1.links(cells[0]);
+    auto dofs0 = dofmap0.links(cell);
+    auto dofs1 = dofmap1.links(cell);
     if (!bc0.empty())
     {
       for (int i = 0; i < num_dofs0; ++i)
@@ -239,16 +230,16 @@ void assemble_exterior_facets(
       }
     }
 
-    mat_set(dofs0.size(), dofs0.data(), dofs1.size(), dofs1.data(), Ae.data());
+    mat_set(dofs0, dofs1, Ae);
   }
 }
 
 /// Execute kernel over interior facets and  accumulate result in Mat
-template <typename T>
+template <typename T, typename U>
 void assemble_interior_facets(
-    const std::function<int(std::int32_t, const std::int32_t*, std::int32_t,
-                            const std::int32_t*, const T*)>& mat_set,
-    const mesh::Mesh& mesh, const xtl::span<const std::int32_t>& active_facets,
+    U mat_set, const mesh::Mesh& mesh,
+    const xtl::span<const std::tuple<std::int32_t, int, std::int32_t, int>>&
+        facets,
     const std::function<void(const xtl::span<T>&,
                              const xtl::span<const std::uint32_t>&,
                              std::int32_t, int)>& dof_transform,
@@ -256,69 +247,60 @@ void assemble_interior_facets(
     const std::function<void(const xtl::span<T>&,
                              const xtl::span<const std::uint32_t>&,
                              std::int32_t, int)>& dof_transform_to_transpose,
-    const DofMap& dofmap1, int bs1, const std::vector<bool>& bc0,
-    const std::vector<bool>& bc1,
+    const DofMap& dofmap1, int bs1, const xtl::span<const std::int8_t>& bc0,
+    const xtl::span<const std::int8_t>& bc1,
     const std::function<void(T*, const T*, const T*, const double*, const int*,
                              const std::uint8_t*)>& kernel,
-    const array2d<T>& coeffs, const xtl::span<const int>& offsets,
-    const xtl::span<const T>& constants,
+    const xtl::span<const T>& coeffs, int cstride,
+    const xtl::span<const int>& offsets, const xtl::span<const T>& constants,
     const xtl::span<const std::uint32_t>& cell_info,
     const std::function<std::uint8_t(std::size_t)>& get_perm)
 {
+  if (facets.empty())
+    return;
+
   const int tdim = mesh.topology().dim();
 
   // Prepare cell geometry
   const graph::AdjacencyList<std::int32_t>& x_dofmap = mesh.geometry().dofmap();
 
-  // FIXME: Add proper interface for num coordinate dofs
-  const std::size_t num_dofs_g = x_dofmap.num_links(0);
-  const xt::xtensor<double, 2>& x_g = mesh.geometry().x();
+  const std::size_t num_dofs_g = mesh.geometry().cmap().dim();
+
+  xtl::span<const double> x_g = mesh.geometry().x();
 
   // Data structures used in assembly
   xt::xtensor<double, 3> coordinate_dofs({2, num_dofs_g, 3});
   std::vector<T> Ae, be;
   std::vector<T> coeff_array(2 * offsets.back());
-  assert(offsets.back() == coeffs.shape[1]);
+  assert(offsets.back() == cstride);
+
+  const int num_cell_facets
+      = mesh::cell_num_entities(mesh.topology().cell_type(), tdim - 1);
 
   // Temporaries for joint dofmaps
   std::vector<std::int32_t> dmapjoint0, dmapjoint1;
 
-  // Iterate over all facets
-  auto c = mesh.topology().connectivity(tdim - 1, tdim);
-  assert(c);
-  auto c_to_f = mesh.topology().connectivity(tdim, tdim - 1);
-  assert(c_to_f);
-
-  for (std::int32_t facet_index : active_facets)
+  for (std::size_t index = 0; index < facets.size(); ++index)
   {
-    // Create attached cells
-    auto cells = c->links(facet_index);
-    assert(cells.size() == 2);
-
-    // Get local index of facet with respect to the cell
-    auto facets0 = c_to_f->links(cells[0]);
-    const auto* it0 = std::find(facets0.begin(), facets0.end(), facet_index);
-    assert(it0 != facets0.end());
-    const int local_facet0 = std::distance(facets0.begin(), it0);
-    auto facets1 = c_to_f->links(cells[1]);
-    const auto* it1 = std::find(facets1.begin(), facets1.end(), facet_index);
-    assert(it1 != facets1.end());
-    const int local_facet1 = std::distance(facets1.begin(), it1);
-
-    const std::array local_facet = {local_facet0, local_facet1};
+    const std::array<std::int32_t, 2> cells
+        = {std::get<0>(facets[index]), std::get<2>(facets[index])};
+    const std::array<int, 2> local_facet
+        = {std::get<1>(facets[index]), std::get<3>(facets[index])};
 
     // Get cell geometry
     auto x_dofs0 = x_dofmap.links(cells[0]);
     for (std::size_t i = 0; i < x_dofs0.size(); ++i)
     {
-      std::copy_n(xt::view(x_g, x_dofs0[i]).begin(), 3,
-                  xt::view(coordinate_dofs, 0, i, xt::all()).begin());
+      common::impl::copy_N<3>(
+          std::next(x_g.begin(), 3 * x_dofs0[i]),
+          xt::view(coordinate_dofs, 0, i, xt::all()).begin());
     }
     auto x_dofs1 = x_dofmap.links(cells[1]);
     for (std::size_t i = 0; i < x_dofs1.size(); ++i)
     {
-      std::copy_n(xt::view(x_g, x_dofs1[i]).begin(), 3,
-                  xt::view(coordinate_dofs, 1, i, xt::all()).begin());
+      common::impl::copy_N<3>(
+          std::next(x_g.begin(), 3 * x_dofs1[i]),
+          xt::view(coordinate_dofs, 1, i, xt::all()).begin());
     }
 
     // Get dof maps for cells and pack
@@ -336,22 +318,6 @@ void assemble_interior_facets(
     std::copy(dmap1_cell1.begin(), dmap1_cell1.end(),
               std::next(dmapjoint1.begin(), dmap1_cell0.size()));
 
-    // Layout for the restricted coefficients is flattened
-    // w[coefficient][restriction][dof]
-    auto coeff_cell0 = coeffs.row(cells[0]);
-    auto coeff_cell1 = coeffs.row(cells[1]);
-
-    // Loop over coefficients
-    for (std::size_t i = 0; i < offsets.size() - 1; ++i)
-    {
-      // Loop over entries for coefficient i
-      const int num_entries = offsets[i + 1] - offsets[i];
-      std::copy_n(coeff_cell0.data() + offsets[i], num_entries,
-                  std::next(coeff_array.begin(), 2 * offsets[i]));
-      std::copy_n(coeff_cell1.data() + offsets[i], num_entries,
-                  std::next(coeff_array.begin(), offsets[i + 1] + offsets[i]));
-    }
-
     const int num_rows = bs0 * dmapjoint0.size();
     const int num_cols = bs1 * dmapjoint1.size();
 
@@ -359,15 +325,30 @@ void assemble_interior_facets(
     Ae.resize(num_rows * num_cols);
     std::fill(Ae.begin(), Ae.end(), 0);
 
-    const int facets_per_cell = facets0.size();
     const std::array perm{
-        get_perm(cells[0] * facets_per_cell + local_facet[0]),
-        get_perm(cells[1] * facets_per_cell + local_facet[1])};
-    kernel(Ae.data(), coeff_array.data(), constants.data(),
+        get_perm(cells[0] * num_cell_facets + local_facet[0]),
+        get_perm(cells[1] * num_cell_facets + local_facet[1])};
+    kernel(Ae.data(), coeffs.data() + index * 2 * cstride, constants.data(),
            coordinate_dofs.data(), local_facet.data(), perm.data());
 
-    dof_transform(Ae, cell_info, cells[0], num_cols);
-    dof_transform_to_transpose(Ae, cell_info, cells[0], num_rows);
+    const xtl::span<T> _Ae(Ae);
+
+    const xtl::span<T> sub_Ae0
+        = _Ae.subspan(bs0 * dmap0_cell0.size() * num_cols,
+                      bs0 * dmap0_cell1.size() * num_cols);
+    const xtl::span<T> sub_Ae1
+        = _Ae.subspan(bs1 * dmap1_cell0.size(),
+                      num_rows * num_cols - bs1 * dmap1_cell0.size());
+
+    // Need to apply DOF transformations for parts of the matrix due to cell 0
+    // and cell 1. For example, if the space has 3 DOFs, then Ae will be 6 by 6
+    // (3 rows/columns for each cell). Subspans are used to offset to the right
+    // blocks of the matrix
+
+    dof_transform(_Ae, cell_info, cells[0], num_cols);
+    dof_transform(sub_Ae0, cell_info, cells[1], num_cols);
+    dof_transform_to_transpose(_Ae, cell_info, cells[0], num_rows);
+    dof_transform_to_transpose(sub_Ae1, cell_info, cells[1], num_rows);
 
     // Zero rows/columns for essential bcs
     if (!bc0.empty())
@@ -401,22 +382,20 @@ void assemble_interior_facets(
       }
     }
 
-    mat_set(dmapjoint0.size(), dmapjoint0.data(), dmapjoint1.size(),
-            dmapjoint1.data(), Ae.data());
+    mat_set(dmapjoint0, dmapjoint1, Ae);
   }
 }
 
-template <typename T>
+template <typename T, typename U>
 void assemble_matrix(
-    const std::function<int(std::int32_t, const std::int32_t*, std::int32_t,
-                            const std::int32_t*, const T*)>& mat_set,
-    const Form<T>& a, const xtl::span<const T>& constants,
-    const array2d<T>& coeffs, const std::vector<bool>& bc0,
-    const std::vector<bool>& bc1)
+    U mat_set, const Form<T>& a, const xtl::span<const T>& constants,
+    const std::map<std::pair<IntegralType, int>,
+                   std::pair<xtl::span<const T>, int>>& coefficients,
+    const xtl::span<const std::int8_t>& bc0,
+    const xtl::span<const std::int8_t>& bc1)
 {
   std::shared_ptr<const mesh::Mesh> mesh = a.mesh();
   assert(mesh);
-  const int tdim = mesh->topology().dim();
 
   // Get dofmap data
   std::shared_ptr<const fem::DofMap> dofmap0
@@ -457,18 +436,28 @@ void assemble_matrix(
   for (int i : a.integral_ids(IntegralType::cell))
   {
     const auto& fn = a.kernel(IntegralType::cell, i);
-    const std::vector<std::int32_t>& active_cells
-        = a.domains(IntegralType::cell, i);
-    impl::assemble_cells<T>(mat_set, mesh->geometry(), active_cells,
-                            dof_transform, dofs0, bs0,
-                            dof_transform_to_transpose, dofs1, bs1, bc0, bc1,
-                            fn, coeffs, constants, cell_info);
+    const auto& [coeffs, cstride] = coefficients.at({IntegralType::cell, i});
+    const std::vector<std::int32_t>& cells = a.cell_domains(i);
+    impl::assemble_cells(mat_set, mesh->geometry(), cells, dof_transform, dofs0,
+                         bs0, dof_transform_to_transpose, dofs1, bs1, bc0, bc1,
+                         fn, coeffs, cstride, constants, cell_info);
   }
 
-  if (a.num_integrals(IntegralType::exterior_facet) > 0
-      or a.num_integrals(IntegralType::interior_facet) > 0)
+  for (int i : a.integral_ids(IntegralType::exterior_facet))
+  {
+    const auto& fn = a.kernel(IntegralType::exterior_facet, i);
+    const auto& [coeffs, cstride]
+        = coefficients.at({IntegralType::exterior_facet, i});
+    const std::vector<std::pair<std::int32_t, int>>& facets
+        = a.exterior_facet_domains(i);
+    impl::assemble_exterior_facets(mat_set, *mesh, facets, dof_transform, dofs0,
+                                   bs0, dof_transform_to_transpose, dofs1, bs1,
+                                   bc0, bc1, fn, coeffs, cstride, constants,
+                                   cell_info);
+  }
+
+  if (a.num_integrals(IntegralType::interior_facet) > 0)
   {
-    mesh->topology_mutable().create_connectivity(tdim - 1, tdim);
     std::function<std::uint8_t(std::size_t)> get_perm;
     if (a.needs_facet_permutations())
     {
@@ -480,27 +469,19 @@ void assemble_matrix(
     else
       get_perm = [](std::size_t) { return 0; };
 
-    for (int i : a.integral_ids(IntegralType::exterior_facet))
-    {
-      const auto& fn = a.kernel(IntegralType::exterior_facet, i);
-      const std::vector<std::int32_t>& active_facets
-          = a.domains(IntegralType::exterior_facet, i);
-      impl::assemble_exterior_facets<T>(
-          mat_set, *mesh, active_facets, dof_transform, dofs0, bs0,
-          dof_transform_to_transpose, dofs1, bs1, bc0, bc1, fn, coeffs,
-          constants, cell_info, get_perm);
-    }
-
     const std::vector<int> c_offsets = a.coefficient_offsets();
     for (int i : a.integral_ids(IntegralType::interior_facet))
     {
       const auto& fn = a.kernel(IntegralType::interior_facet, i);
-      const std::vector<std::int32_t>& active_facets
-          = a.domains(IntegralType::interior_facet, i);
-      impl::assemble_interior_facets<T>(
-          mat_set, *mesh, active_facets, dof_transform, *dofmap0, bs0,
+      const auto& [coeffs, cstride]
+          = coefficients.at({IntegralType::interior_facet, i});
+      const std::vector<std::tuple<std::int32_t, int, std::int32_t, int>>&
+          facets
+          = a.interior_facet_domains(i);
+      impl::assemble_interior_facets(
+          mat_set, *mesh, facets, dof_transform, *dofmap0, bs0,
           dof_transform_to_transpose, *dofmap1, bs1, bc0, bc1, fn, coeffs,
-          c_offsets, constants, cell_info, get_perm);
+          cstride, c_offsets, constants, cell_info, get_perm);
     }
   }
 }
