@@ -7,6 +7,7 @@
 #include "dofmapbuilder.h"
 #include "ElementDofLayout.h"
 #include <algorithm>
+#include <cstdint>
 #include <cstdlib>
 #include <dolfinx/common/IndexMap.h>
 #include <dolfinx/common/MPI.h>
@@ -14,11 +15,11 @@
 #include <dolfinx/common/utils.h>
 #include <dolfinx/graph/AdjacencyList.h>
 #include <dolfinx/mesh/Topology.h>
+#include <dolfinx/mesh/cell_types.h>
 #include <iterator>
 #include <memory>
 #include <numeric>
 #include <random>
-#include <unordered_map>
 #include <utility>
 #include <vector>
 
@@ -138,7 +139,7 @@ build_basic_dofmap(const mesh::Topology& topology,
   const int D = topology.dim();
 
   // Generate and number required mesh entities
-  std::vector<bool> needs_entities(D + 1, false);
+  std::vector<std::int8_t> needs_entities(D + 1, false);
   std::vector<std::int32_t> num_mesh_entities_local(D + 1, 0),
       num_mesh_entities_global(D + 1, 0);
   for (int d = 0; d <= D; ++d)
@@ -203,7 +204,7 @@ build_basic_dofmap(const mesh::Topology& topology,
   }
 
   // Entity dofs on cell (dof = entity_dofs[dim][entity][index])
-  const std::vector<std::vector<std::set<int>>>& entity_dofs
+  const std::vector<std::vector<std::vector<int>>>& entity_dofs
       = element_dof_layout.entity_dofs_all();
 
   // Storage for local-to-global map
@@ -238,7 +239,6 @@ build_basic_dofmap(const mesh::Topology& topology,
 
     // Iterate over each topological dimension
     std::int32_t offset_local = 0;
-    std::int64_t offset_global = 0;
     for (auto e_dofs_d = entity_dofs.begin(); e_dofs_d != entity_dofs.end();
          ++e_dofs_d)
     {
@@ -251,7 +251,6 @@ build_basic_dofmap(const mesh::Topology& topology,
         // global)
         const std::size_t e = std::distance(e_dofs_d->begin(), e_dofs);
         const std::int32_t e_index_local = entity_indices_local[d][e];
-        const std::int64_t e_index_global = entity_indices_global[d][e];
 
         // Loop over dofs belong to entity e of dimension d (d, e)
         // d: topological dimension
@@ -265,13 +264,39 @@ build_basic_dofmap(const mesh::Topology& topology,
           const std::int32_t dof
               = offset_local + num_entity_dofs * e_index_local + count;
           dofs[cell_ptr[c] + *dof_local] = dof;
+        }
+      }
+      offset_local += entity_dofs[d][0].size() * num_mesh_entities_local[d];
+    }
+  }
+
+  // Create local to global map and dof entity map. NOTE this must be done
+  // outside of the above loop as some processes may have vertices that don't
+  // belong to a cell on that process.
+  std::int32_t offset_local = 0;
+  std::int64_t offset_global = 0;
+  for (int d = 0; d <= D; ++d)
+  {
+    if (needs_entities[d])
+    {
+      // NOTE This assumes all entities have the same number of dofs
+      auto num_entity_dofs = entity_dofs[d][0].size();
+      for (std::int32_t e_index_local = 0;
+           e_index_local < num_mesh_entities_local[d]; ++e_index_local)
+      {
+        auto e_index_global = global_indices[d][e_index_local];
+
+        for (std::size_t count = 0; count < num_entity_dofs; ++count)
+        {
+          const std::int32_t dof
+              = offset_local + num_entity_dofs * e_index_local + count;
           local_to_global[dof]
               = offset_global + num_entity_dofs * e_index_global + count;
           dof_entity[dof] = {d, e_index_local};
         }
       }
-      offset_local += entity_dofs[d][0].size() * num_mesh_entities_local[d];
-      offset_global += entity_dofs[d][0].size() * num_mesh_entities_global[d];
+      offset_local += num_entity_dofs * num_mesh_entities_local[d];
+      offset_global += num_entity_dofs * num_mesh_entities_global[d];
     }
   }
 
@@ -342,6 +367,21 @@ std::pair<std::vector<std::int32_t>, std::int32_t> compute_reordering_map(
     }
   }
 
+  // Check for any -1's remaining in `original_to_contiguous` due to vertices
+  // on the process that don't belong to a cell. Determine if the dof is owned
+  // or a ghost and map to the ends of the owned and ghost "parts" of the
+  // contiguous array respectively.
+  for (std::size_t dof = 0; dof < original_to_contiguous.size(); ++dof)
+  {
+    if (original_to_contiguous[dof] == -1)
+    {
+      if (auto e = dof_entity[dof]; e.second < offset[e.first])
+        original_to_contiguous[dof] = counter_owned++;
+      else
+        original_to_contiguous[dof] = counter_unowned++;
+    }
+  }
+
   if (reorder_fn)
   {
     // Re-order using graph ordering
@@ -349,9 +389,10 @@ std::pair<std::vector<std::int32_t>, std::int32_t> compute_reordering_map(
     // Apply graph reordering to owned dofs
     const std::vector<int> node_remap
         = reorder_owned(dofmap, owned_size, original_to_contiguous, reorder_fn);
-    std::for_each(original_to_contiguous.begin(), original_to_contiguous.end(),
-                  [&node_remap, owned_size](auto index)
-                  { return index < owned_size ? node_remap[index] : index; });
+    std::transform(original_to_contiguous.begin(), original_to_contiguous.end(),
+                   original_to_contiguous.begin(),
+                   [&node_remap, owned_size](auto index)
+                   { return index < owned_size ? node_remap[index] : index; });
   }
 
   return {std::move(original_to_contiguous), owned_size};
@@ -384,13 +425,13 @@ std::pair<std::vector<std::int64_t>, std::vector<int>> get_global_indices(
 
   // Build list flag for owned mesh entities that are shared, i.e. are a
   // ghost on a neighbor
-  std::vector<std::vector<bool>> shared_entity(D + 1);
+  std::vector<std::vector<std::int8_t>> shared_entity(D + 1);
   for (std::size_t d = 0; d < shared_entity.size(); ++d)
   {
     auto map = topology.index_map(d);
     if (map)
     {
-      shared_entity[d] = std::vector<bool>(map->size_local(), false);
+      shared_entity[d] = std::vector<std::int8_t>(map->size_local(), false);
       const std::vector<std::int32_t>& forward_indices
           = map->scatter_fwd_indices().array();
       for (auto entity : forward_indices)
@@ -485,21 +526,31 @@ std::pair<std::vector<std::int64_t>, std::vector<int>> get_global_indices(
     auto [neighbors, _] = dolfinx::MPI::neighbors(comm[d]);
 
     // Build (global old, global new) map for dofs of dimension d
-    std::unordered_map<std::int64_t, std::pair<int64_t, int>> global_old_new;
+    std::vector<std::pair<std::int64_t, std::pair<int64_t, int>>>
+        global_old_new;
+    global_old_new.reserve(disp_recv[d].back());
     for (std::size_t j = 0; j < all_dofs_received[d].size(); j += 2)
     {
       const auto pos
           = std::upper_bound(disp_recv[d].begin(), disp_recv[d].end(), j);
       const int owner = std::distance(disp_recv[d].begin(), pos) - 1;
-      global_old_new.insert({all_dofs_received[d][j],
-                             {all_dofs_received[d][j + 1], neighbors[owner]}});
+      global_old_new.push_back(
+          {all_dofs_received[d][j],
+           {all_dofs_received[d][j + 1], neighbors[owner]}});
     }
+    std::sort(global_old_new.begin(), global_old_new.end());
 
     // Build the dimension d part of local_to_global_new vector
     for (std::size_t i = 0; i < local_new_to_global_old[d].size(); i += 2)
     {
-      auto it = global_old_new.find(local_new_to_global_old[d][i]);
-      assert(it != global_old_new.end());
+      std::pair<std::int64_t, std::pair<int64_t, int>> idx_old
+          = {local_new_to_global_old[d][i], {0, 0}};
+
+      auto it = std::lower_bound(
+          global_old_new.begin(), global_old_new.end(), idx_old,
+          [](auto& a, auto& b) { return a.first < b.first; });
+      assert(it != global_old_new.end() and it->first == idx_old.first);
+
       local_to_global_new[local_new_to_global_old[d][i + 1]] = it->second.first;
       local_to_global_new_owner[local_new_to_global_old[d][i + 1]]
           = it->second.second;
@@ -512,8 +563,7 @@ std::pair<std::vector<std::int64_t>, std::vector<int>> get_global_indices(
 } // namespace
 
 //-----------------------------------------------------------------------------
-std::tuple<std::shared_ptr<common::IndexMap>, int,
-           graph::AdjacencyList<std::int32_t>>
+std::tuple<common::IndexMap, int, graph::AdjacencyList<std::int32_t>>
 fem::build_dofmap_data(
     MPI_Comm comm, const mesh::Topology& topology,
     const ElementDofLayout& element_dof_layout,
@@ -531,15 +581,13 @@ fem::build_dofmap_data(
   const auto [node_graph0, local_to_global0, dof_entity0]
       = build_basic_dofmap(topology, element_dof_layout);
 
-  // Compute global dofmap dimension
-  std::int64_t global_dimension(0), offset(0);
+  // Compute global dofmap offset
+  std::int64_t offset = 0;
   for (int d = 0; d <= D; ++d)
   {
     if (element_dof_layout.num_entity_dofs(d) > 0)
     {
       assert(topology.index_map(d));
-      const std::int64_t n = topology.index_map(d)->size_global();
-      global_dimension += n * element_dof_layout.num_entity_dofs(d);
       offset += topology.index_map(d)->local_range()[0]
                 * element_dof_layout.num_entity_dofs(d);
     }
@@ -557,13 +605,13 @@ fem::build_dofmap_data(
   assert(local_to_global_unowned.size() == local_to_global_owner.size());
 
   // Create IndexMap for dofs range on this process
-  auto index_map = std::make_unique<common::IndexMap>(
-      comm, num_owned,
-      dolfinx::MPI::compute_graph_edges(
-          comm, std::set<int>(local_to_global_owner.begin(),
-                              local_to_global_owner.end())),
-      local_to_global_unowned, local_to_global_owner);
-  assert(index_map);
+  std::vector<int> src_ranks = local_to_global_owner;
+  std::sort(src_ranks.begin(), src_ranks.end());
+  src_ranks.erase(std::unique(src_ranks.begin(), src_ranks.end()),
+                  src_ranks.end());
+  auto dest_ranks = dolfinx::MPI::compute_graph_edges_nbx(comm, src_ranks);
+  common::IndexMap index_map(comm, num_owned, dest_ranks,
+                             local_to_global_unowned, local_to_global_owner);
 
   // Build re-ordered dofmap
   std::vector<std::int32_t> dofmap(node_graph0.array().size());
@@ -580,9 +628,12 @@ fem::build_dofmap_data(
     }
   }
 
-  assert(dofmap.size() % node_graph0.num_nodes() == 0);
+  int dofs_per_cell
+      = dofmap.empty() ? 0 : dofmap.size() / node_graph0.num_nodes();
+  graph::AdjacencyList<std::int32_t> dmap
+      = graph::regular_adjacency_list(std::move(dofmap), dofs_per_cell);
+
   return {std::move(index_map), element_dof_layout.block_size(),
-          graph::build_adjacency_list<std::int32_t>(
-              std::move(dofmap), dofmap.size() / node_graph0.num_nodes())};
+          std::move(dmap)};
 }
 //-----------------------------------------------------------------------------
