@@ -7,7 +7,6 @@
 #include "xdmf_utils.h"
 #include "pugixml.hpp"
 #include <boost/algorithm/string.hpp>
-#include <boost/filesystem.hpp>
 #include <boost/lexical_cast.hpp>
 #include <dolfinx/common/IndexMap.h>
 #include <dolfinx/fem/CoordinateElement.h>
@@ -21,23 +20,112 @@
 #include <dolfinx/mesh/Topology.h>
 #include <dolfinx/mesh/cell_types.h>
 #include <dolfinx/mesh/utils.h>
+#include <filesystem>
 #include <map>
 #include <xtensor/xadapt.hpp>
 #include <xtensor/xtensor.hpp>
-#include <xtensor/xview.hpp>
 
 using namespace dolfinx;
 using namespace dolfinx::io;
 
 namespace
 {
+/// @warning Do not use. This function will be removed.
+///
+/// Send in_values[p0] to process p0 and receive values from process p1
+/// in out_values[p1]
+template <typename T>
+graph::AdjacencyList<T> all_to_all(MPI_Comm comm,
+                                   const graph::AdjacencyList<T>& send_data)
+{
+  const std::vector<std::int32_t>& send_offsets = send_data.offsets();
+  const std::vector<T>& values_in = send_data.array();
+
+  const int comm_size = dolfinx::MPI::size(comm);
+  assert(send_data.num_nodes() == comm_size);
+
+  // Data size per destination rank
+  std::vector<int> send_size(comm_size);
+  std::adjacent_difference(std::next(send_offsets.begin()), send_offsets.end(),
+                           send_size.begin());
+
+  // Get received data sizes from each rank
+  std::vector<int> recv_size(comm_size);
+  MPI_Alltoall(send_size.data(), 1, MPI_INT, recv_size.data(), 1, MPI_INT,
+               comm);
+
+  // Compute receive offset
+  std::vector<std::int32_t> recv_offset(comm_size + 1, 0);
+  std::partial_sum(recv_size.begin(), recv_size.end(),
+                   std::next(recv_offset.begin()));
+
+  // Send/receive data
+  std::vector<T> recv_values(recv_offset.back());
+  MPI_Alltoallv(values_in.data(), send_size.data(), send_offsets.data(),
+                dolfinx::MPI::mpi_type<T>(), recv_values.data(),
+                recv_size.data(), recv_offset.data(),
+                dolfinx::MPI::mpi_type<T>(), comm);
+
+  return graph::AdjacencyList<T>(std::move(recv_values),
+                                 std::move(recv_offset));
+}
+
+/// Compute values at all mesh 'nodes'
+/// @return The values at all geometric points
+/// @warning This function will be removed soon. Use interpolation
+/// instead.
+template <typename T>
+xt::xtensor<T, 2> compute_point_values(const fem::Function<T>& u)
+{
+  auto V = u.function_space();
+  assert(V);
+  std::shared_ptr<const mesh::Mesh> mesh = V->mesh();
+  assert(mesh);
+  const int tdim = mesh->topology().dim();
+
+  // Compute in tensor (one for scalar function, . . .)
+  const std::size_t value_size_loc = V->element()->value_size();
+
+  // Resize Array for holding point values
+  xt::xtensor<T, 2> point_values
+      = xt::zeros<T>({mesh->geometry().x().size() / 3, value_size_loc});
+
+  // Prepare cell geometry
+  const graph::AdjacencyList<std::int32_t>& x_dofmap
+      = mesh->geometry().dofmap();
+  const int num_dofs_g = mesh->geometry().cmap().dim();
+  const auto x_g
+      = xt::adapt(mesh->geometry().x().data(), mesh->geometry().x().size(),
+                  xt::no_ownership(),
+                  std::vector{mesh->geometry().x().size() / 3, std::size_t(3)});
+
+  // Interpolate point values on each cell (using last computed value if
+  // not continuous, e.g. discontinuous Galerkin methods)
+  auto map = mesh->topology().index_map(tdim);
+  assert(map);
+  const std::int32_t num_cells = map->size_local() + map->num_ghosts();
+
+  std::vector<std::int32_t> cells(x_g.shape(0), -1);
+  for (std::int32_t c = 0; c < num_cells; ++c)
+  {
+    // Get coordinates for all points in cell
+    xtl::span<const std::int32_t> dofs = x_dofmap.links(c);
+    for (int i = 0; i < num_dofs_g; ++i)
+      cells[dofs[i]] = c;
+  }
+
+  u.eval(x_g, cells, point_values);
+
+  return point_values;
+}
+
 //-----------------------------------------------------------------------------
 // Get data width - normally the same as u.value_size(), but expand for
 // 2D vector/tensor because XDMF presents everything as 3D
 std::int64_t get_padded_width(const fem::FiniteElement& e)
 {
   const int width = e.value_size();
-  const int rank = e.value_rank();
+  const int rank = e.value_shape().size();
   if (rank == 1 and width == 2)
     return 3;
   else if (rank == 2 and width == 4)
@@ -51,7 +139,7 @@ std::vector<Scalar> _get_point_data_values(const fem::Function<Scalar>& u)
 {
   std::shared_ptr<const mesh::Mesh> mesh = u.function_space()->mesh();
   assert(mesh);
-  const xt::xtensor<Scalar, 2> data_values = u.compute_point_values();
+  const xt::xtensor<Scalar, 2> data_values = compute_point_values(u);
 
   const int width = get_padded_width(*u.function_space()->element());
   assert(mesh->geometry().index_map());
@@ -62,7 +150,7 @@ std::vector<Scalar> _get_point_data_values(const fem::Function<Scalar>& u)
   // FIXME: Unpick the below code for the new layout of data from
   //        GenericFunction::compute_vertex_values
   std::vector<Scalar> _data_values(width * num_local_points, 0.0);
-  const int value_rank = u.function_space()->element()->value_rank();
+  const int value_rank = u.function_space()->element()->value_shape().size();
   if (value_rank > 0)
   {
     // Transpose vector/tensor data arrays
@@ -93,7 +181,7 @@ std::vector<Scalar> _get_cell_data_values(const fem::Function<Scalar>& u)
   assert(u.function_space()->dofmap());
   const auto mesh = u.function_space()->mesh();
   const int value_size = u.function_space()->element()->value_size();
-  const int value_rank = u.function_space()->element()->value_rank();
+  const int value_rank = u.function_space()->element()->value_shape().size();
 
   // Allocate memory for function values at cell centres
   const int tdim = mesh->topology().dim();
@@ -105,8 +193,7 @@ std::vector<Scalar> _get_cell_data_values(const fem::Function<Scalar>& u)
   std::vector<std::int32_t> dof_set;
   dof_set.reserve(local_size);
   const auto dofmap = u.function_space()->dofmap();
-  assert(dofmap->element_dof_layout);
-  const int ndofs = dofmap->element_dof_layout->num_dofs();
+  const int ndofs = dofmap->element_dof_layout().num_dofs();
   const int bs = dofmap->bs();
   assert(ndofs * bs == value_size);
 
@@ -123,7 +210,7 @@ std::vector<Scalar> _get_cell_data_values(const fem::Function<Scalar>& u)
 
   // Get values
   std::vector<Scalar> values(dof_set.size());
-  const std::vector<Scalar>& _u = u.x()->array();
+  xtl::span<const Scalar> _u = u.x()->array();
   for (std::size_t i = 0; i < dof_set.size(); ++i)
     values[i] = _u[dof_set[i]];
 
@@ -233,17 +320,18 @@ xdmf_utils::get_hdf5_paths(const pugi::xml_node& dataitem_node)
   return {{paths[0], paths[1]}};
 }
 //-----------------------------------------------------------------------------
-std::string xdmf_utils::get_hdf5_filename(std::string xdmf_filename)
+std::filesystem::path
+xdmf_utils::get_hdf5_filename(const std::filesystem::path& xdmf_filename)
 {
-  boost::filesystem::path p(xdmf_filename);
-  p.replace_extension(".h5");
+  std::filesystem::path p = xdmf_filename;
+  p.replace_extension("h5");
   if (p.string() == xdmf_filename)
   {
     throw std::runtime_error("Cannot deduce name of HDF5 file from XDMF "
                              "filename. Filename clash. Check XDMF filename");
   }
 
-  return p.string();
+  return p;
 }
 //-----------------------------------------------------------------------------
 std::vector<std::int64_t>
@@ -337,6 +425,7 @@ std::string xdmf_utils::vtk_cell_type_str(mesh::CellType cell_type,
        {{4, "Quadrilateral"},
         {9, "Quadrilateral_9"},
         {16, "Quadrilateral_16"}}},
+      {mesh::CellType::prism, {{6, "Wedge"}}},
       {mesh::CellType::tetrahedron,
        {{4, "Tetrahedron"}, {10, "Tetrahedron_10"}, {20, "Tetrahedron_20"}}},
       {mesh::CellType::hexahedron, {{8, "Hexahedron"}, {27, "Hexahedron_27"}}}};
@@ -354,63 +443,29 @@ std::string xdmf_utils::vtk_cell_type_str(mesh::CellType cell_type,
   return cell_str->second;
 }
 //-----------------------------------------------------------------------------
-std::pair<xt::xtensor<std::int32_t, 2>, std::vector<std::int32_t>>
-xdmf_utils::extract_local_entities(const mesh::Mesh& mesh, const int entity_dim,
-                                   const xt::xtensor<std::int64_t, 2>& entities,
-                                   const xtl::span<const std::int32_t>& values)
+std::pair<std::vector<std::int32_t>, std::vector<std::int32_t>>
+xdmf_utils::distribute_entity_data(
+    const mesh::Mesh& mesh, int entity_dim,
+    const xtl::span<const std::int64_t>& entities,
+    const xtl::span<const std::int32_t>& data)
 {
-  if (entities.shape(0) != values.size())
-    throw std::runtime_error("Number of entities and values must match");
-
-  // Get layout of dofs on 0th entity
-  const std::vector<int> entity_layout
-      = mesh.geometry().cmap().dof_layout().entity_closure_dofs(entity_dim, 0);
-  assert(entity_layout.size() == entities.shape(1));
-
-  auto c_to_v = mesh.topology().connectivity(mesh.topology().dim(), 0);
-  if (!c_to_v)
-    throw std::runtime_error("Missing cell-vertex connectivity.");
+  LOG(INFO) << "XDMF distribute entity data";
 
   // Use ElementDofLayout of the cell to get vertex dof indices (local
-  // to a cell) i.e. find a map from local vertex index to associated
+  // to a cell), i.e. build a map from local vertex index to associated
   // local dof index
-  const int num_vertices_per_cell = c_to_v->num_links(0);
-  std::vector<int> cell_vertex_dofs(num_vertices_per_cell);
-  for (int i = 0; i < num_vertices_per_cell; ++i)
+  std::vector<int> cell_vertex_dofs;
   {
-    const std::vector<int> local_index
-        = mesh.geometry().cmap().dof_layout().entity_dofs(0, i);
-    assert(local_index.size() == 1);
-    cell_vertex_dofs[i] = local_index[0];
-  }
-
-  // Find map from entity vertex to local (wrt. dof numbering on the
-  // entity) dof number E.g. if there are dofs on entity [0 3 6 7 9] and
-  // dofs 3 and 7 belong to vertices, then this produces map [1, 3]
-  std::vector<int> entity_vertex_dofs;
-  for (std::size_t i = 0; i < cell_vertex_dofs.size(); ++i)
-  {
-    auto it = std::find(entity_layout.begin(), entity_layout.end(),
-                        cell_vertex_dofs[i]);
-    if (it != entity_layout.end())
-      entity_vertex_dofs.push_back(std::distance(entity_layout.begin(), it));
-  }
-
-  const mesh::CellType entity_type
-      = mesh::cell_entity_type(mesh.topology().cell_type(), entity_dim);
-  const std::size_t num_vertices_per_entity
-      = mesh::cell_num_entities(entity_type, 0);
-  assert(entity_vertex_dofs.size() == num_vertices_per_entity);
-
-  // Throw away input global indices which do not belong to entity
-  // vertices This decreases the amount of data needed in parallel
-  // communication
-  xt::xtensor<std::int64_t, 2> entities_vertices(
-      {entities.shape(0), num_vertices_per_entity});
-  for (std::size_t e = 0; e < entities_vertices.shape(0); ++e)
-  {
-    for (std::size_t i = 0; i < entities_vertices.shape(1); ++i)
-      entities_vertices(e, i) = entities(e, entity_vertex_dofs[i]);
+    // Get layout of dofs on 0th cell entity of dimension entity_dim
+    const fem::ElementDofLayout cmap_dof_layout
+        = mesh.geometry().cmap().create_dof_layout();
+    for (int i = 0; i < mesh::cell_num_entities(mesh.topology().cell_type(), 0);
+         ++i)
+    {
+      const std::vector<int>& local_index = cmap_dof_layout.entity_dofs(0, i);
+      assert(local_index.size() == 1);
+      cell_vertex_dofs.push_back(local_index[0]);
+    }
   }
 
   // -------------------
@@ -418,31 +473,41 @@ xdmf_utils::extract_local_entities(const mesh::Mesh& mesh, const int entity_dim,
   //    'postmaster' rank, and receive global "input" nodes for which
   //    this rank is the postmaster
 
-  // Get "input" global node indices (as in the input file before any
-  // internal re-ordering)
-  const std::vector<std::int64_t>& nodes_g
-      = mesh.geometry().input_global_indices();
-
-  // Send input global indices to 'post master' rank, based on input
-  // global index value
-  const std::int64_t num_nodes_g = mesh.geometry().index_map()->size_global();
-  const MPI_Comm comm = mesh.mpi_comm();
-  const int comm_size = MPI::size(comm);
-  // NOTE: could make this int32_t be sending: index <- index - dest_rank_offset
-  std::vector<std::vector<std::int64_t>> nodes_g_send(comm_size);
-  for (std::int64_t node : nodes_g)
+  auto postmaster_global_nodes_sendrecv = [](const mesh::Mesh& mesh)
   {
-    // TODO: Optimise this call by adding 'vectorised verion of
-    //       MPI::index_owner
-    // Figure out which process is the postmaster for the input global index
-    const std::int32_t p
-        = dolfinx::MPI::index_owner(comm_size, node, num_nodes_g);
-    nodes_g_send[p].push_back(node);
-  }
+    const MPI_Comm comm = mesh.comm();
+    const int comm_size = dolfinx::MPI::size(comm);
+
+    // Get "input" global node indices (as in the input file before any
+    // internal re-ordering)
+    const std::vector<std::int64_t>& nodes_g
+        = mesh.geometry().input_global_indices();
+
+    // Send input global indices to 'post master' rank, based on input
+    // global index value
+    const std::int64_t num_nodes_g = mesh.geometry().index_map()->size_global();
+    // NOTE: could make this int32_t be sending: index <- index -
+    // dest_rank_offset
+    std::vector<std::vector<std::int64_t>> nodes_g_send(comm_size);
+    for (std::int64_t node : nodes_g)
+    {
+      // Figure out which process is the postmaster for the input global
+      // index
+      const std::int32_t p
+          = dolfinx::MPI::index_owner(comm_size, node, num_nodes_g);
+      nodes_g_send[p].push_back(node);
+    }
+
+    // Send/receive
+    LOG(INFO) << "XDMF send entity nodes size:(" << num_nodes_g << ")";
+    graph::AdjacencyList<std::int64_t> nodes_g_recv
+        = all_to_all(comm, graph::AdjacencyList<std::int64_t>(nodes_g_send));
+
+    return nodes_g_recv;
+  };
 
-  // Send/receive
   const graph::AdjacencyList<std::int64_t> nodes_g_recv
-      = MPI::all_to_all(comm, graph::AdjacencyList<std::int64_t>(nodes_g_send));
+      = postmaster_global_nodes_sendrecv(mesh);
 
   // -------------------
   // 2. Send the entity key (nodes list) and tag to the postmaster based
@@ -452,29 +517,81 @@ xdmf_utils::extract_local_entities(const mesh::Mesh& mesh, const int entity_dim,
   //    data (i) the communication could be combined, or (ii) the
   //    communication in Step 1 could be make non-blocking.
 
-  std::vector<std::vector<std::int64_t>> entities_send(comm_size);
-  std::vector<std::vector<std::int32_t>> values_send(comm_size);
-  std::vector<std::int64_t> entity(num_vertices_per_entity);
-  for (std::size_t e = 0; e < entities_vertices.shape(0); ++e)
+  auto postmaster_global_ent_sendrecv
+      = [&cell_vertex_dofs](const mesh::Mesh& mesh, int entity_dim,
+                            const xtl::span<const std::int64_t>& entities,
+                            const xtl::span<const std::int32_t>& data)
   {
-    // Copy vertices for entity and sort
-    auto ev = xt::row(entities_vertices, e);
-    std::copy(ev.cbegin(), ev.cend(), entity.begin());
-    std::sort(entity.begin(), entity.end());
+    const MPI_Comm comm = mesh.comm();
+    const int comm_size = dolfinx::MPI::size(comm);
+    const std::int64_t num_nodes_g = mesh.geometry().index_map()->size_global();
 
-    // Determine postmaster based on lowest entity node
-    const std::int32_t p
-        = dolfinx::MPI::index_owner(comm_size, entity.front(), num_nodes_g);
-    entities_send[p].insert(entities_send[p].end(), entity.begin(),
-                            entity.end());
-    values_send[p].push_back(values[e]);
-  }
+    const std::size_t num_vert_per_entity = mesh::cell_num_entities(
+        mesh::cell_entity_type(mesh.topology().cell_type(), entity_dim, 0), 0);
+    auto c_to_v = mesh.topology().connectivity(mesh.topology().dim(), 0);
+    if (!c_to_v)
+      throw std::runtime_error("Missing cell-vertex connectivity.");
 
-  // TODO: Pack into one MPI call
-  const graph::AdjacencyList<std::int64_t> entities_recv = MPI::all_to_all(
-      comm, graph::AdjacencyList<std::int64_t>(entities_send));
-  const graph::AdjacencyList<std::int32_t> values_recv
-      = MPI::all_to_all(comm, graph::AdjacencyList<std::int32_t>(values_send));
+    const fem::ElementDofLayout cmap_dof_layout
+        = mesh.geometry().cmap().create_dof_layout();
+    const std::vector<int> entity_layout
+        = cmap_dof_layout.entity_closure_dofs(entity_dim, 0);
+
+    // Find map from entity vertex to local (w.r.t. dof numbering on the
+    // entity) dof number. E.g., if there are dofs on entity [0 3 6 7 9]
+    // and dofs 3 and 7 belong to vertices, then this produces map [1,
+    // 3].
+    std::vector<int> entity_vertex_dofs;
+    for (std::size_t i = 0; i < cell_vertex_dofs.size(); ++i)
+    {
+      auto it = std::find(entity_layout.begin(), entity_layout.end(),
+                          cell_vertex_dofs[i]);
+      if (it != entity_layout.end())
+        entity_vertex_dofs.push_back(std::distance(entity_layout.begin(), it));
+    }
+
+    const std::size_t shape_e_1 = entity_layout.size();
+    const std::size_t shape_e_0 = entities.size() / shape_e_1;
+    std::vector<std::int64_t> entities_vertices(shape_e_0
+                                                * num_vert_per_entity);
+    for (std::size_t e = 0; e < shape_e_0; ++e)
+    {
+      for (std::size_t i = 0; i < num_vert_per_entity; ++i)
+      {
+        entities_vertices[e * num_vert_per_entity + i]
+            = entities[e * shape_e_1 + entity_vertex_dofs[i]];
+      }
+    }
+
+    std::vector<std::vector<std::int64_t>> entities_send(comm_size);
+    std::vector<std::vector<std::int32_t>> data_send(comm_size);
+    for (std::size_t e = 0; e < shape_e_0; ++e)
+    {
+      xtl::span<std::int64_t> entity(entities_vertices.data()
+                                         + e * num_vert_per_entity,
+                                     num_vert_per_entity);
+      std::sort(entity.begin(), entity.end());
+
+      // Determine postmaster based on lowest entity node
+      const std::int32_t p
+          = dolfinx::MPI::index_owner(comm_size, entity.front(), num_nodes_g);
+      entities_send[p].insert(entities_send[p].end(), entity.begin(),
+                              entity.end());
+      data_send[p].push_back(data[e]);
+    }
+
+    LOG(INFO) << "XDMF send entity keys size: (" << shape_e_0 << ")";
+    // TODO: Pack into one MPI call
+    graph::AdjacencyList<std::int64_t> entities_recv
+        = all_to_all(comm, graph::AdjacencyList<std::int64_t>(entities_send));
+    graph::AdjacencyList<std::int32_t> data_recv
+        = all_to_all(comm, graph::AdjacencyList<std::int32_t>(data_send));
+
+    return std::pair(entities_recv, data_recv);
+  };
+
+  const auto [entities_recv, data_recv]
+      = postmaster_global_ent_sendrecv(mesh, entity_dim, entities, data);
 
   // -------------------
   // 3. As 'postmaster', send back the entity key (vertex list) and tag
@@ -485,48 +602,67 @@ xdmf_utils::extract_local_entities(const mesh::Mesh& mesh, const int entity_dim,
   // owned nodes to the postmaster and use map, unordered_map or
   // std::vector<pair>>, followed by a neighborhood all_to_all at the
   // end.
-  //
-  // Build map from global node index to ranks that have the node
-  std::multimap<std::int64_t, int> node_to_rank;
-  for (int p = 0; p < nodes_g_recv.num_nodes(); ++p)
+
+  auto postmaster_send_to_candidates
+      = [](const mesh::Mesh& mesh, int entity_dim,
+           const graph::AdjacencyList<std::int64_t>& nodes_g_recv,
+           const graph::AdjacencyList<std::int64_t>& entities_recv,
+           const graph::AdjacencyList<std::int32_t>& data_recv)
   {
-    auto nodes = nodes_g_recv.links(p);
-    for (std::int32_t node : nodes)
-      node_to_rank.insert({node, p});
-  }
+    const MPI_Comm comm = mesh.comm();
+    const int comm_size = dolfinx::MPI::size(comm);
 
-  // Figure out which processes are owners of received nodes
-  std::vector<std::vector<std::int64_t>> send_nodes_owned(comm_size);
-  std::vector<std::vector<std::int32_t>> send_vals_owned(comm_size);
-  std::array<std::size_t, 2> shape
-      = {entities_recv.array().size() / num_vertices_per_entity,
-         num_vertices_per_entity};
-  auto _entities_recv
-      = xt::adapt(entities_recv.array().data(), entities_recv.array().size(),
-                  xt::no_ownership(), shape);
+    const std::size_t num_vert_per_entity = mesh::cell_num_entities(
+        mesh::cell_entity_type(mesh.topology().cell_type(), entity_dim, 0), 0);
 
-  const std::vector<std::int32_t>& _values_recv = values_recv.array();
-  assert(_values_recv.size() == _entities_recv.shape(0));
-  for (std::size_t e = 0; e < _entities_recv.shape(0); ++e)
-  {
-    auto e_recv = xt::row(_entities_recv, e);
-
-    // Find ranks that have node0
-    auto [it0, it1] = node_to_rank.equal_range(e_recv[0]);
-    for (auto it = it0; it != it1; ++it)
+    // Build map from global node index to ranks that have the node
+    std::multimap<std::int64_t, int> node_to_rank;
+    for (int p = 0; p < nodes_g_recv.num_nodes(); ++p)
     {
-      const int p1 = it->second;
-      send_nodes_owned[p1].insert(send_nodes_owned[p1].end(), e_recv.begin(),
-                                  e_recv.end());
-      send_vals_owned[p1].push_back(_values_recv[e]);
+      auto nodes = nodes_g_recv.links(p);
+      for (std::int32_t node : nodes)
+        node_to_rank.insert({node, p});
     }
-  }
 
-  // TODO: Pack into one MPI call
-  const graph::AdjacencyList<std::int64_t> recv_ents = MPI::all_to_all(
-      comm, graph::AdjacencyList<std::int64_t>(send_nodes_owned));
-  const graph::AdjacencyList<std::int32_t> recv_vals = MPI::all_to_all(
-      comm, graph::AdjacencyList<std::int32_t>(send_vals_owned));
+    // Figure out which processes are owners of received nodes
+    std::vector<std::vector<std::int64_t>> send_nodes_owned(comm_size);
+    std::vector<std::vector<std::int32_t>> send_vals_owned(comm_size);
+    const std::size_t shape0
+        = entities_recv.array().size() / num_vert_per_entity;
+    const std::size_t shape1 = num_vert_per_entity;
+    const std::vector<std::int32_t>& _data_recv = data_recv.array();
+    assert(_data_recv.size() == shape0);
+    for (std::size_t e = 0; e < shape0; ++e)
+    {
+      xtl::span e_recv(entities_recv.array().data() + e * shape1, shape1);
+
+      // Find ranks that have node0
+      auto [it0, it1] = node_to_rank.equal_range(e_recv.front());
+      for (auto it = it0; it != it1; ++it)
+      {
+        const int p1 = it->second;
+        send_nodes_owned[p1].insert(send_nodes_owned[p1].end(), e_recv.begin(),
+                                    e_recv.end());
+        send_vals_owned[p1].push_back(_data_recv[e]);
+      }
+    }
+
+    // TODO: Pack into one MPI call
+    const int send_val_size = std::transform_reduce(
+        send_vals_owned.begin(), send_vals_owned.end(), 0, std::plus<int>(),
+        [](const std::vector<std::int32_t>& v) { return v.size(); });
+    LOG(INFO) << "XDMF return entity and value data size:(" << send_val_size
+              << ")";
+    graph::AdjacencyList<std::int64_t> recv_ents = all_to_all(
+        comm, graph::AdjacencyList<std::int64_t>(send_nodes_owned));
+    graph::AdjacencyList<std::int32_t> recv_vals
+        = all_to_all(comm, graph::AdjacencyList<std::int32_t>(send_vals_owned));
+
+    return std::pair(std::move(recv_ents), std::move(recv_vals));
+  };
+
+  const auto [recv_ents, recv_vals] = postmaster_send_to_candidates(
+      mesh, entity_dim, nodes_g_recv, entities_recv, data_recv);
 
   // -------------------
   // 4. From the received (key, value) data, determine which keys
@@ -542,57 +678,72 @@ xdmf_utils::extract_local_entities(const mesh::Mesh& mesh, const int entity_dim,
   //       the std::map for *all* entities and just for candidate
   //       entities.
 
-  // Build map from input global indices to local vertex numbers
-  const graph::AdjacencyList<std::int32_t>& x_dofmap = mesh.geometry().dofmap();
-  std::map<std::int64_t, std::int32_t> igi_to_vertex;
-
-  for (int c = 0; c < c_to_v->num_nodes(); ++c)
+  auto determine_my_entities
+      = [&cell_vertex_dofs](const mesh::Mesh& mesh, int entity_dim,
+                            const graph::AdjacencyList<std::int64_t>& recv_ents,
+                            const graph::AdjacencyList<std::int32_t>& recv_vals)
   {
-    auto vertices = c_to_v->links(c);
-    auto x_dofs = x_dofmap.links(c);
-    for (std::size_t v = 0; v < vertices.size(); ++v)
-      igi_to_vertex[nodes_g[x_dofs[cell_vertex_dofs[v]]]] = vertices[v];
-  }
+    // Build map from input global indices to local vertex numbers
+    LOG(INFO) << "XDMF build map";
 
-  std::vector<std::int32_t> entities_new;
-  entities_new.reserve(recv_ents.array().size());
-  std::vector<std::int32_t> values_new;
-  values_new.reserve(recv_vals.array().size());
-  for (std::size_t e = 0;
-       e < recv_ents.array().size() / num_vertices_per_entity; ++e)
-  {
-    bool entity_found = true;
-    std::vector<std::int32_t> entity(num_vertices_per_entity);
-    for (std::size_t i = 0; i < num_vertices_per_entity; ++i)
+    const std::size_t num_vert_per_entity = mesh::cell_num_entities(
+        mesh::cell_entity_type(mesh.topology().cell_type(), entity_dim, 0), 0);
+    auto c_to_v = mesh.topology().connectivity(mesh.topology().dim(), 0);
+    if (!c_to_v)
+      throw std::runtime_error("Missing cell-vertex connectivity.");
+
+    const std::vector<std::int64_t>& nodes_g
+        = mesh.geometry().input_global_indices();
+
+    const graph::AdjacencyList<std::int32_t>& x_dofmap
+        = mesh.geometry().dofmap();
+    std::map<std::int64_t, std::int32_t> igi_to_vertex;
+    for (int c = 0; c < c_to_v->num_nodes(); ++c)
     {
-      const auto it = igi_to_vertex.find(
-          recv_ents.array()[e * num_vertices_per_entity + i]);
-      if (it == igi_to_vertex.end())
+      auto vertices = c_to_v->links(c);
+      auto x_dofs = x_dofmap.links(c);
+      for (std::size_t v = 0; v < vertices.size(); ++v)
+        igi_to_vertex[nodes_g[x_dofs[cell_vertex_dofs[v]]]] = vertices[v];
+    }
+
+    std::vector<std::int32_t> entities_new;
+    entities_new.reserve(recv_ents.array().size());
+    std::vector<std::int32_t> data_new;
+    data_new.reserve(recv_vals.array().size());
+    std::vector<std::int32_t> entity(num_vert_per_entity);
+    const std::vector<std::int64_t>& recv_ents_array = recv_ents.array();
+    for (std::size_t e = 0; e < recv_ents_array.size() / num_vert_per_entity;
+         ++e)
+    {
+      bool entity_found = true;
+      for (std::size_t i = 0; i < num_vert_per_entity; ++i)
       {
-        // As soon as this received index is not in locally owned input
-        // global indices skip the entire entity
-        entity_found = false;
-        break;
+        if (auto it
+            = igi_to_vertex.find(recv_ents_array[e * num_vert_per_entity + i]);
+            it == igi_to_vertex.end())
+        {
+          // As soon as this received index is not in locally owned input
+          // global indices skip the entire entity
+          entity_found = false;
+          break;
+        }
+        else
+          entity[i] = it->second;
+      }
+
+      if (entity_found == true)
+      {
+        entities_new.insert(entities_new.end(), entity.begin(), entity.end());
+        data_new.push_back(recv_vals.array()[e]);
       }
-      entity[i] = it->second;
     }
 
-    if (entity_found == true)
-    {
-      entities_new.insert(entities_new.end(), entity.begin(), entity.end());
-      values_new.push_back(recv_vals.array()[e]);
-    }
-  }
+    return std::pair(std::move(entities_new), std::move(data_new));
+  };
 
-  std::array<std::size_t, 2> shape_r = {
-      entities_new.size() / num_vertices_per_entity, num_vertices_per_entity};
+  auto [entities_new, data_new]
+      = determine_my_entities(mesh, entity_dim, recv_ents, recv_vals);
 
-  // The below should work, but misbehaves with the Intel icpx compiler
-  // auto e_new = xt::adapt(entities_new.data(), entities_new.size(),
-  //                        xt::no_ownership(), shape_r);
-  xt::xtensor<std::int32_t, 2> e_new(shape_r);
-  std::copy_n(entities_new.data(), entities_new.size(), e_new.data());
-
-  return {std::move(e_new), std::move(values_new)};
+  return {std::move(entities_new), std::move(data_new)};
 }
 //-----------------------------------------------------------------------------
