@@ -15,7 +15,6 @@
 #include <dolfinx/common/Timer.h>
 #include <dolfinx/common/log.h>
 #include <dolfinx/common/sort.h>
-#include <dolfinx/common/utils.h>
 #include <dolfinx/graph/AdjacencyList.h>
 #include <memory>
 #include <numeric>
@@ -25,8 +24,6 @@
 #include <unordered_map>
 #include <utility>
 #include <vector>
-#include <xtensor/xtensor.hpp>
-#include <xtensor/xview.hpp>
 
 using namespace dolfinx;
 
@@ -37,10 +34,11 @@ namespace
 /// @param processes Set of sharing processes
 /// @param vertices Global vertex indices of entity
 /// @return owning process number
-int get_ownership(std::set<int>& processes, std::vector<std::int64_t>& vertices)
+int get_ownership(const std::set<int>& processes,
+                  const std::vector<std::int64_t>& vertices)
 {
-  // Use a deterministic random number generator, seeded with global vertex
-  // indices ensuring all processes get the same answer
+  // Use a deterministic random number generator, seeded with global
+  // vertex indices ensuring all processes get the same answer
   std::mt19937 gen;
   std::seed_seq seq(vertices.begin(), vertices.end());
   gen.seed(seq);
@@ -62,15 +60,17 @@ int get_ownership(std::set<int>& processes, std::vector<std::int64_t>& vertices)
 /// @param[in] entity_list List of entities, each entity represented by
 /// its local vertex indices
 /// @param[in] num_vertices_per_e Number of vertices per entity
+/// @param[in] num_entities_per_cell Number of entities per cell
 /// @param[in] entity_index Initial numbering for each row in
 /// entity_list
 /// @returns Tuple of (local_indices, index map, shared entities)
-std::tuple<std::vector<int>, std::shared_ptr<common::IndexMap>>
-get_local_indexing(
-    MPI_Comm comm, const std::shared_ptr<const common::IndexMap>& cell_indexmap,
-    const std::shared_ptr<const common::IndexMap>& vertex_indexmap,
-    const xt::xtensor<std::int32_t, 2>& entity_list,
-    const std::vector<std::int32_t>& entity_index)
+std::tuple<std::vector<int>, common::IndexMap>
+get_local_indexing(MPI_Comm comm, const common::IndexMap& cell_indexmap,
+                   const common::IndexMap& vertex_indexmap,
+                   const xtl::span<const std::int32_t>& entity_list,
+                   std::int32_t num_vertices_per_e,
+                   std::int32_t num_entities_per_cell,
+                   const xtl::span<const std::int32_t>& entity_index)
 {
   // entity_list contains all the entities for all the cells,
   // listed as local vertex indices, and entity_index contains the
@@ -94,21 +94,16 @@ get_local_indexing(
   // Set ghost status array values
   // 1 = entities that are only in local cells (i.e. owned)
   // 2 = entities that are only in ghost cells (i.e. not owned)
-  // 3 = entities with ownership that needs deciding (used also for unghosted
-  // case)
+  // 3 = entities with ownership that needs deciding (used also for
+  // un-ghosted case)
   std::vector<int> ghost_status(entity_count, 0);
   {
-    if (cell_indexmap->num_ghosts() == 0)
+    if (cell_indexmap.num_ghosts() == 0)
       std::fill(ghost_status.begin(), ghost_status.end(), 3);
     else
     {
-      const std::int32_t num_cells
-          = cell_indexmap->size_local() + cell_indexmap->num_ghosts();
-      assert(entity_list.shape(0) % num_cells == 0);
-      const std::int32_t num_entities_per_cell
-          = entity_list.shape(0) / num_cells;
       const std::int32_t ghost_offset
-          = cell_indexmap->size_local() * num_entities_per_cell;
+          = cell_indexmap.size_local() * num_entities_per_cell;
 
       // Tag all entities in local cells with 1
       for (int i = 0; i < ghost_offset; ++i)
@@ -118,7 +113,7 @@ get_local_indexing(
       }
 
       // Set entities in ghost cells to 2 (purely ghost) or 3 (border)
-      for (std::size_t i = ghost_offset; i < entity_list.shape(0); ++i)
+      for (std::size_t i = ghost_offset; i < entity_index.size(); ++i)
       {
         const std::int32_t idx = entity_index[i];
         ghost_status[idx] = ghost_status[idx] | 2;
@@ -129,7 +124,7 @@ get_local_indexing(
   //---------
   // Create an expanded neighbor_comm from shared_vertices
   const std::map<std::int32_t, std::set<std::int32_t>> shared_vertices
-      = vertex_indexmap->compute_shared_indices();
+      = vertex_indexmap.compute_shared_indices();
 
   std::set<std::int32_t> neighbor_set;
   for (auto& q : shared_vertices)
@@ -159,21 +154,19 @@ get_local_indexing(
   // Set of sharing procs for each entity, counting vertex hits
   // Get a list of unique entities
   std::vector<std::int32_t> unique_row(entity_count);
-  for (std::size_t i = 0; i < entity_list.shape(0); ++i)
+  for (std::size_t i = 0; i < entity_index.size(); ++i)
     unique_row[entity_index[i]] = i;
-  const int num_vertices_per_e = entity_list.shape(1);
   std::unordered_map<int, int> procs;
   std::vector<std::int64_t> vglobal(num_vertices_per_e);
-  std::vector<std::int32_t> entity_list_i(num_vertices_per_e);
   for (int i : unique_row)
   {
-    std::copy_n(xt::row(entity_list, i).begin(), num_vertices_per_e,
-                entity_list_i.begin());
+    xtl::span entity_list_i
+        = entity_list.subspan(i * num_vertices_per_e, num_vertices_per_e);
+
     procs.clear();
-    for (int j = 0; j < num_vertices_per_e; ++j)
+    for (auto v : entity_list_i)
     {
-      if (auto it = shared_vertices.find(entity_list_i[j]);
-          it != shared_vertices.end())
+      if (auto it = shared_vertices.find(v); it != shared_vertices.end())
       {
         for (std::int32_t p : it->second)
           ++procs[p];
@@ -185,7 +178,7 @@ get_local_indexing(
       // If any process shares all vertices, then add to list
       if (q.second == num_vertices_per_e)
       {
-        vertex_indexmap->local_to_global(entity_list_i, vglobal);
+        vertex_indexmap.local_to_global(entity_list_i, vglobal);
         dolfinx::radix_sort(xtl::span(vglobal));
 
         global_entity_to_entity_index.insert({vglobal, entity_index[i]});
@@ -302,8 +295,7 @@ get_local_indexing(
   {
     const std::int64_t _num_local = num_local;
     std::int64_t local_offset = 0;
-    MPI_Exscan(&_num_local, &local_offset, 1,
-               dolfinx::MPI::mpi_type<std::int64_t>(), MPI_SUM, comm);
+    MPI_Exscan(&_num_local, &local_offset, 1, MPI_INT64_T, MPI_SUM, comm);
 
     std::vector<std::int64_t> send_global_index_data;
     std::vector<int> send_global_index_offsets = {0};
@@ -357,12 +349,13 @@ get_local_indexing(
   }
 
   MPI_Comm_free(&neighbor_comm);
-
-  auto index_map = std::make_shared<common::IndexMap>(
-      comm, num_local,
-      dolfinx::MPI::compute_graph_edges(
-          comm, std::set<int>(ghost_owners.begin(), ghost_owners.end())),
-      ghost_indices, ghost_owners);
+  std::vector<int> src_ranks = ghost_owners;
+  std::sort(src_ranks.begin(), src_ranks.end());
+  src_ranks.erase(std::unique(src_ranks.begin(), src_ranks.end()),
+                  src_ranks.end());
+  auto dest_ranks = dolfinx::MPI::compute_graph_edges_nbx(comm, src_ranks);
+  common::IndexMap index_map(comm, num_local, dest_ranks, ghost_indices,
+                             ghost_owners);
 
   // Map from initial numbering to new local indices
   std::vector<std::int32_t> new_entity_index(entity_index.size());
@@ -370,27 +363,26 @@ get_local_indexing(
                  new_entity_index.begin(),
                  [&local_index](auto index) { return local_index[index]; });
 
-  return {std::move(new_entity_index), index_map};
+  return {std::move(new_entity_index), std::move(index_map)};
 }
 //-----------------------------------------------------------------------------
 
 /// Compute entities of dimension d
+///
 /// @param[in] comm MPI communicator (TODO: full or neighbor hood?)
 /// @param[in] cells Adjacency list for cell-vertex connectivity
 /// @param[in] shared_vertices TODO
 /// @param[in] cell_type Cell type
 /// @param[in] dim Topological dimension of the entities to be computed
 /// @return Returns the (cell-entity connectivity, entity-cell
-///   connectivity, index map for the entity distribution across
-///   processes, shared entities)
-std::tuple<std::shared_ptr<graph::AdjacencyList<std::int32_t>>,
-           std::shared_ptr<graph::AdjacencyList<std::int32_t>>,
-           std::shared_ptr<common::IndexMap>>
+/// connectivity, index map for the entity distribution across
+/// processes, shared entities)
+std::tuple<graph::AdjacencyList<std::int32_t>,
+           graph::AdjacencyList<std::int32_t>, common::IndexMap>
 compute_entities_by_key_matching(
     MPI_Comm comm, const graph::AdjacencyList<std::int32_t>& cells,
-    const std::shared_ptr<const common::IndexMap>& vertex_index_map,
-    const std::shared_ptr<const common::IndexMap>& cell_index_map,
-    mesh::CellType cell_type, int dim)
+    const common::IndexMap& vertex_index_map,
+    const common::IndexMap& cell_index_map, mesh::CellType cell_type, int dim)
 {
   if (dim == 0)
   {
@@ -402,75 +394,119 @@ compute_entities_by_key_matching(
   common::Timer timer("Compute entities of dim = " + std::to_string(dim));
 
   // Initialize local array of entities
-  const std::int8_t num_entities_per_cell
-      = mesh::cell_num_entities(cell_type, dim);
-  const std::size_t num_vertices_per_entity
-      = mesh::num_cell_vertices(mesh::cell_entity_type(cell_type, dim));
+  const std::int8_t num_entities_per_cell = cell_num_entities(cell_type, dim);
+
+  // For some cells, the num_vertices varies per facet (3 or 4)
+  int max_vertices_per_entity = 0;
+  for (int i = 0; i < num_entities_per_cell; ++i)
+  {
+    max_vertices_per_entity
+        = std::max(max_vertices_per_entity,
+                   num_cell_vertices(cell_entity_type(cell_type, dim, i)));
+  }
 
   // Create map from cell vertices to entity vertices
-  auto e_vertices = mesh::get_entity_vertices(cell_type, dim);
+  auto e_vertices = get_entity_vertices(cell_type, dim);
 
   // List of vertices for each entity in each cell
   const std::size_t num_cells = cells.num_nodes();
-  xt::xtensor<std::int32_t, 2> entity_list(
-      {num_cells * num_entities_per_cell, num_vertices_per_entity});
+  const std::size_t entity_list_shape0 = num_cells * num_entities_per_cell;
+  const std::size_t entity_list_shape1 = max_vertices_per_entity;
+  std::vector<std::int32_t> entity_list(entity_list_shape0 * entity_list_shape1,
+                                        -1);
   for (std::size_t c = 0; c < num_cells; ++c)
   {
     // Get vertices from cell
     auto vertices = cells.links(c);
+
     for (int i = 0; i < num_entities_per_cell; ++i)
     {
       const std::int32_t idx = c * num_entities_per_cell + i;
       auto ev = e_vertices.links(i);
 
-      // Get entity vertices
-      assert(e_vertices.num_links(i) == (int)num_vertices_per_entity);
-      for (std::size_t j = 0; j < num_vertices_per_entity; ++j)
-        entity_list(idx, j) = vertices[ev[j]];
+      // Get entity vertices. Padded with -1 if fewer than
+      // max_vertices_per_entity
+      for (std::size_t j = 0; j < ev.size(); ++j)
+        entity_list[idx * entity_list_shape1 + j] = vertices[ev[j]];
     }
   }
 
-  // Copy list and sort vertices of each entity into order
-  xt::xtensor<std::int32_t, 2> entity_list_sorted = entity_list;
-  for (std::size_t i = 0; i < entity_list_sorted.shape(0); ++i)
-  {
-    std::sort(xt::row(entity_list_sorted, i).begin(),
-              xt::row(entity_list_sorted, i).end());
-  }
-
-  // Sort the list and label uniquely
-  const std::vector<std::int32_t> sort_order
-      = dolfinx::sort_by_perm(entity_list_sorted);
-
-  std::vector<std::int32_t> entity_index(entity_list.shape(0), 0);
+  std::vector<std::int32_t> entity_index(entity_list_shape0, -1);
   std::int32_t entity_count = 0;
-  std::int32_t last = sort_order[0];
-  for (std::size_t i = 1; i < sort_order.size(); ++i)
   {
-    std::int32_t j = sort_order[i];
-    if (xt::row(entity_list_sorted, j) != xt::row(entity_list_sorted, last))
+    // Copy list and sort vertices of each entity into (reverse) order
+    std::vector<std::int32_t> entity_list_sorted = entity_list;
+    for (std::size_t i = 0; i < entity_list_shape0; ++i)
+    {
+      auto it = std::next(entity_list_sorted.begin(), i * entity_list_shape1);
+      std::sort(it, std::next(it, entity_list_shape1), std::greater<>());
+    }
+
+    // Sort the list and label uniquely
+    const std::vector<std::int32_t> sort_order
+        = dolfinx::sort_by_perm<std::int32_t>(entity_list_sorted,
+                                              entity_list_shape1);
+
+    std::vector<std::int32_t> entity(max_vertices_per_entity),
+        entity0(max_vertices_per_entity);
+    auto it = sort_order.begin();
+    while (it != sort_order.end())
+    {
+      // First entity in new index range
+      std::size_t offset = (*it) * max_vertices_per_entity;
+      xtl::span e0(entity_list_sorted.data() + offset, max_vertices_per_entity);
+
+      // Find iterator to next entity
+      auto it1 = std::find_if_not(
+          it, sort_order.end(),
+          [e0, &entity_list_sorted, max_vertices_per_entity](auto idx) -> bool
+          {
+            std::size_t offset = idx * max_vertices_per_entity;
+            return std::equal(e0.cbegin(), e0.cend(),
+                              std::next(entity_list_sorted.begin(), offset));
+          });
+
+      // Set entity unique index
+      std::for_each(it, it1,
+                    [&entity_index, entity_count](auto idx)
+                    { entity_index[idx] = entity_count; });
+
+      // Advance iterator and increment entity
+      it = it1;
       ++entity_count;
-    entity_index[j] = entity_count;
-    last = j;
+    }
   }
-  ++entity_count;
 
   // Communicate with other processes to find out which entities are
   // ghosted and shared. Remap the numbering so that ghosts are at the
   // end.
-  const auto [local_index, index_map] = get_local_indexing(
-      comm, cell_index_map, vertex_index_map, entity_list, entity_index);
+  auto [local_index, index_map] = get_local_indexing(
+      comm, cell_index_map, vertex_index_map, entity_list,
+      max_vertices_per_entity, num_entities_per_cell, entity_index);
 
   // Entity-vertex connectivity
   std::vector<std::int32_t> offsets_ev(entity_count + 1, 0);
-  for (std::size_t i = 0; i < offsets_ev.size() - 1; ++i)
-    offsets_ev[i + 1] = offsets_ev[i] + num_vertices_per_entity;
-  auto ev = std::make_shared<graph::AdjacencyList<std::int32_t>>(
-      std::vector<std::int32_t>(offsets_ev.back()), std::move(offsets_ev));
-  for (std::size_t i = 0; i < entity_list.shape(0); ++i)
+  std::vector<int> size_ev(entity_count);
+  for (std::size_t i = 0; i < entity_list_shape0; ++i)
   {
-    std::copy(xt::row(entity_list, i).begin(), xt::row(entity_list, i).end(),
-              ev->links(local_index[i]).begin());
+    // if (entity_list(i, max_vertices_per_entity - 1) == -1)
+    if (entity_list[i * entity_list_shape1 + entity_list_shape1 - 1] == -1)
+      size_ev[local_index[i]] = max_vertices_per_entity - 1;
+    else
+      size_ev[local_index[i]] = max_vertices_per_entity;
+  }
+
+  std::transform(size_ev.cbegin(), size_ev.cend(), offsets_ev.cbegin(),
+                 std::next(offsets_ev.begin()),
+                 [](auto a, auto b) { return a + b; });
+
+  graph::AdjacencyList<std::int32_t> ev(
+      std::vector<std::int32_t>(offsets_ev.back()), std::move(offsets_ev));
+  for (std::size_t i = 0; i < entity_list_shape0; ++i)
+  {
+    auto _ev = ev.links(local_index[i]);
+    std::copy_n(std::next(entity_list.begin(), i * entity_list_shape1),
+                _ev.size(), _ev.begin());
   }
 
   // NOTE: Cell-entity connectivity comes after ev creation because
@@ -478,22 +514,25 @@ compute_entities_by_key_matching(
 
   // Cell-entity connectivity
   std::vector<std::int32_t> offsets_ce(num_cells + 1, 0);
-  for (std::size_t i = 0; i < offsets_ce.size() - 1; ++i)
-    offsets_ce[i + 1] = offsets_ce[i] + num_entities_per_cell;
-  auto ce = std::make_shared<graph::AdjacencyList<std::int32_t>>(
-      std::move(local_index), std::move(offsets_ce));
+  std::transform(offsets_ce.cbegin(), std::prev(offsets_ce.cend()),
+                 std::next(offsets_ce.begin()),
+                 [num_entities_per_cell](auto x)
+                 { return x + num_entities_per_cell; });
+  graph::AdjacencyList<std::int32_t> ce(std::move(local_index),
+                                        std::move(offsets_ce));
 
-  return {ce, ev, index_map};
+  return {std::move(ce), std::move(ev), std::move(index_map)};
 }
 //-----------------------------------------------------------------------------
 
 /// Compute connectivity from entities of dimension d0 to entities of
 /// dimension d1 using the transpose connectivity (d1 -> d0)
+///
 /// @param[in] c_d1_d0 The connectivity from entities of dimension d1 to
-///   entities of dimension d0
-/// @param[in] num_entities_d0 The number of entities of dimension d0
+/// entities of dimension d0.
+/// @param[in] num_entities_d0 The number of entities of dimension d0.
 /// @return The connectivity from entities of dimension d0 to entities
-///   of dimension d1
+/// of dimension d1.
 graph::AdjacencyList<std::int32_t>
 compute_from_transpose(const graph::AdjacencyList<std::int32_t>& c_d1_d0,
                        const int num_entities_d0, int d0, int d1)
@@ -534,61 +573,51 @@ compute_from_transpose(const graph::AdjacencyList<std::int32_t>& c_d1_d0,
 /// @return The d0 -> d1 connectivity
 graph::AdjacencyList<std::int32_t>
 compute_from_map(const graph::AdjacencyList<std::int32_t>& c_d0_0,
-                 const graph::AdjacencyList<std::int32_t>& c_d1_0,
-                 mesh::CellType cell_type_d0, int d0, int d1)
+                 const graph::AdjacencyList<std::int32_t>& c_d1_0, int d0,
+                 int d1)
 {
-  assert(d1 > 0);
-  assert(d0 > d1);
+  // Only possible case is facet->edge
+  assert(d0 == 2 and d1 == 1);
 
-  // Make a map from the sorted d1 entity vertices to the d1 entity
-  // index
-  boost::unordered_map<std::vector<std::int32_t>, std::int32_t> entity_to_index;
-  entity_to_index.reserve(c_d1_0.num_nodes());
+  // Make a map from the sorted edge vertices to the edge index
+  boost::unordered_map<std::array<std::int32_t, 2>, std::int32_t> edge_to_index;
+  edge_to_index.reserve(c_d1_0.num_nodes());
 
-  const std::size_t num_verts_d1
-      = mesh::num_cell_vertices(mesh::cell_entity_type(cell_type_d0, d1));
-  std::vector<std::int32_t> key(num_verts_d1);
+  std::array<std::int32_t, 2> key;
   for (int e = 0; e < c_d1_0.num_nodes(); ++e)
   {
     xtl::span<const std::int32_t> v = c_d1_0.links(e);
     assert(v.size() == key.size());
     std::partial_sort_copy(v.begin(), v.end(), key.begin(), key.end());
-    entity_to_index.insert({key, e});
+    edge_to_index.insert({key, e});
   }
 
+  // Number of edges for a tri/quad is the same as number of vertices so
+  // AdjacencyList will have same offset pattern
   std::vector<std::int32_t> connections;
-  connections.reserve(c_d0_0.num_nodes()
-                      * mesh::cell_num_entities(cell_type_d0, d1));
-  std::vector<std::int32_t> offsets(c_d0_0.num_nodes() + 1, 0);
+  connections.reserve(c_d0_0.array().size());
+  std::vector<std::int32_t> offsets(c_d0_0.offsets());
+
+  // Search for edges of facet in map, and recover index
+  const auto tri_vertices_ref
+      = get_entity_vertices(mesh::CellType::triangle, 1);
+  const auto quad_vertices_ref
+      = get_entity_vertices(mesh::CellType::quadrilateral, 1);
 
-  // Search for d1 entities of d0 in map, and recover index
-  const auto e_vertices_ref = mesh::get_entity_vertices(cell_type_d0, d1);
-  std::vector<int> keys(e_vertices_ref.array().size());
   for (int e = 0; e < c_d0_0.num_nodes(); ++e)
   {
     auto e0 = c_d0_0.links(e);
-    for (int i = 0; i < e_vertices_ref.num_nodes(); ++i)
+    auto vref = (e0.size() == 3) ? &tri_vertices_ref : &quad_vertices_ref;
+    for (std::size_t i = 0; i < e0.size(); ++i)
     {
-      for (int j = 0; j < e_vertices_ref.num_links(i); ++j)
-      {
-        keys[i * e_vertices_ref.num_links(i) + j]
-            = e0[e_vertices_ref.links(i)[j]];
-      }
-    }
-
-    for (int i = 0; i < e_vertices_ref.num_nodes(); ++i)
-    {
-      auto keys_begin
-          = std::next(keys.cbegin(), i * e_vertices_ref.num_links(i));
-      auto keys_end
-          = std::next(keys.cbegin(), (i + 1) * e_vertices_ref.num_links(i));
-      std::partial_sort_copy(keys_begin, keys_end, key.begin(), key.end());
-      const auto it = entity_to_index.find(key);
-      assert(it != entity_to_index.end());
+      const auto& v = vref->links(i);
+      for (int j = 0; j < 2; ++j)
+        key[j] = e0[v[j]];
+      std::sort(key.begin(), key.end());
+      const auto it = edge_to_index.find(key);
+      assert(it != edge_to_index.end());
       connections.push_back(it->second);
     }
-
-    offsets[e + 1] = offsets[e] + e_vertices_ref.num_nodes();
   }
 
   connections.shrink_to_fit();
@@ -633,13 +662,12 @@ mesh::compute_entities(MPI_Comm comm, const Topology& topology, int dim)
   assert(vertex_map);
   auto cell_map = topology.index_map(tdim);
   assert(cell_map);
-  std::tuple<std::shared_ptr<graph::AdjacencyList<std::int32_t>>,
-             std::shared_ptr<graph::AdjacencyList<std::int32_t>>,
-             std::shared_ptr<common::IndexMap>>
-      data = compute_entities_by_key_matching(
-          comm, *cells, vertex_map, cell_map, topology.cell_type(), dim);
+  auto [d0, d1, d2] = compute_entities_by_key_matching(
+      comm, *cells, *vertex_map, *cell_map, topology.cell_type(), dim);
 
-  return data;
+  return {std::make_shared<graph::AdjacencyList<std::int32_t>>(std::move(d0)),
+          std::make_shared<graph::AdjacencyList<std::int32_t>>(std::move(d1)),
+          std::make_shared<common::IndexMap>(std::move(d2))};
 }
 //-----------------------------------------------------------------------------
 std::array<std::shared_ptr<graph::AdjacencyList<std::int32_t>>, 2>
@@ -685,9 +713,7 @@ mesh::compute_connectivity(const Topology& topology, int d0, int d1)
     if (!topology.connectivity(d1, d0))
     {
       auto c_d1_d0 = std::make_shared<graph::AdjacencyList<std::int32_t>>(
-          compute_from_map(*c_d1_0, *c_d0_0,
-                           mesh::cell_entity_type(topology.cell_type(), d1), d1,
-                           d0));
+          compute_from_map(*c_d1_0, *c_d0_0, d1, d0));
       auto c_d0_d1 = std::make_shared<graph::AdjacencyList<std::int32_t>>(
           compute_from_transpose(*c_d1_d0, c_d0_0->num_nodes(), d0, d1));
       return {c_d0_d1, c_d1_d0};
@@ -706,10 +732,8 @@ mesh::compute_connectivity(const Topology& topology, int d0, int d1)
   {
     // Compute by mapping vertices from a lower dimension entity to
     // those of a higher dimension entity
-    auto c_d0_d1
-        = std::make_shared<graph::AdjacencyList<std::int32_t>>(compute_from_map(
-            *c_d0_0, *c_d1_0, mesh::cell_entity_type(topology.cell_type(), d0),
-            d0, d1));
+    auto c_d0_d1 = std::make_shared<graph::AdjacencyList<std::int32_t>>(
+        compute_from_map(*c_d0_0, *c_d1_0, d0, d1));
     return {c_d0_d1, nullptr};
   }
   else
