@@ -1,4 +1,4 @@
-// Copyright (C) 2010 Garth N. Wells
+// Copyright (C) 2010-2021 Garth N. Wells
 //
 // This file is part of DOLFINx (https://www.fenicsproject.org)
 //
@@ -7,36 +7,37 @@
 #pragma once
 
 #include <cstdint>
+#include <vector>
+#include <xtl/xspan.hpp>
 
 namespace dolfinx::mesh
 {
 class Mesh;
-template <typename T>
-class MeshTags;
 } // namespace dolfinx::mesh
 
 namespace dolfinx::refinement
 {
 
-/// Create uniformly refined mesh
+/// Create a uniformly refined mesh
 ///
 /// @param[in] mesh The mesh from which to build a refined Mesh
 /// @param[in] redistribute Optional argument to redistribute the
-///     refined mesh if mesh is a distributed mesh.
+/// refined mesh if mesh is a distributed mesh.
 /// @return A refined mesh
 mesh::Mesh refine(const mesh::Mesh& mesh, bool redistribute = true);
 
-/// Create locally refined mesh
+/// Create a locally refined mesh
 ///
 /// @param[in] mesh The mesh from which to build a refined Mesh
-/// @param[in] cell_markers A mesh function over integers specifying
-///     which cells should be refined (value == 1) (and which should not
-///     (any other integer value)).
+/// @param[in] edges Indices of the edges that should be split by this
+/// refinement. mesh::compute_incident_entities can be used to compute
+/// the edges that are incident to other entities, e.g. incident to
+/// cells.
 /// @param[in] redistribute Optional argument to redistribute the
-///     refined mesh if mesh is a distributed mesh.
+/// refined mesh if mesh is a distributed mesh.
 /// @return A locally refined mesh
 mesh::Mesh refine(const mesh::Mesh& mesh,
-                  const mesh::MeshTags<std::int8_t>& cell_markers,
+                  const xtl::span<const std::int32_t>& edges,
                   bool redistribute = true);
 
 } // namespace dolfinx::refinement
