@@ -5,28 +5,29 @@
 # SPDX-License-Identifier:    LGPL-3.0-or-later
 """ Unit-tests for higher order meshes """
 
-import os
 import random
+from pathlib import Path
 
 import numpy as np
 import pytest
+
 import ufl
 from dolfinx.cpp.io import perm_gmsh, perm_vtk
-from dolfinx.cpp.mesh import CellType
-from dolfinx.fem import assemble_scalar
+from dolfinx.fem import assemble_scalar, form
 from dolfinx.io import XDMFFile, ufl_mesh_from_gmsh
-from dolfinx.mesh import create_mesh
-from dolfinx_utils.test.skips import skip_in_parallel
-from mpi4py import MPI
+from dolfinx.mesh import CellType, create_mesh
 from ufl import dx
 
+from mpi4py import MPI
+
 
 def check_cell_volume(points, cell, domain, volume):
     random.seed(13)
 
     point_order = [i for i, _ in enumerate(points)]
     for repeat in range(5):
-        # Shuffle the cell to check that permutations of CoordinateElement are correct
+        # Shuffle the cell to check that permutations of
+        # CoordinateElement are correct
         random.shuffle(point_order)
         ordered_points = np.zeros((len(points), len(points[0])))
         for i, j in enumerate(point_order):
@@ -34,12 +35,12 @@ def check_cell_volume(points, cell, domain, volume):
         ordered_cell = [point_order[i] for i in cell]
 
         mesh = create_mesh(MPI.COMM_WORLD, [ordered_cell], ordered_points, domain)
-        area = assemble_scalar(1 * dx(mesh))
+        area = assemble_scalar(form(1 * dx(mesh)))
 
         assert np.isclose(area, volume)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', range(1, 5))
 def test_triangle_mesh(order):
     points = []
@@ -71,7 +72,7 @@ def test_triangle_mesh(order):
     check_cell_volume(points, cell, domain, 0.5)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', range(1, 5))
 def test_tetrahedron_mesh(order):
     points = []
@@ -130,7 +131,7 @@ def test_tetrahedron_mesh(order):
     check_cell_volume(points, cell, domain, 1 / 6)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', [1, 2, 3, 4])
 def test_quadrilateral_mesh(order):
     random.seed(13)
@@ -167,7 +168,7 @@ def test_quadrilateral_mesh(order):
     check_cell_volume(points, cell, domain, 1)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', [1, 2, 3, 4])
 def test_hexahedron_mesh(order):
     random.seed(13)
@@ -245,7 +246,7 @@ def test_hexahedron_mesh(order):
     check_cell_volume(points, cell, domain, 1)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', range(1, 5))
 def test_triangle_mesh_vtk(order):
     points = []
@@ -257,7 +258,8 @@ def test_triangle_mesh_vtk(order):
     def coord_to_vertex(x, y):
         return y * (2 * order + 3 - y) // 2 + x
 
-    # Make the cell, following https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/  # noqa: E501
+    # Make the cell, following
+    # https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/
     cell = [coord_to_vertex(i, j) for i, j in [(0, 0), (order, 0), (0, order)]]
     if order > 1:
         for i in range(1, order):
@@ -284,7 +286,7 @@ def test_triangle_mesh_vtk(order):
     check_cell_volume(points, cell, domain, 0.5)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', range(1, 5))
 def test_tetrahedron_mesh_vtk(order):
     if order > 3:
@@ -303,7 +305,8 @@ def test_tetrahedron_mesh_vtk(order):
             3 * order ** 2 - 3 * order * z + 12 * order + z ** 2 - 6 * z + 11
         ) // 6 + y * (2 * (order - z) + 3 - y) // 2 + x
 
-    # Make the cell, following https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/  # noqa: E501
+    # Make the cell, following
+    # https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/
     cell = [coord_to_vertex(x, y, z) for x, y, z in [
         (0, 0, 0), (order, 0, 0), (0, order, 0), (0, 0, order)]]
 
@@ -322,8 +325,8 @@ def test_tetrahedron_mesh_vtk(order):
             cell.append(coord_to_vertex(0, order - i, i))
 
         if order == 3:
-            # The ordering of faces does not match documentation.
-            # See https://gitlab.kitware.com/vtk/vtk/uploads/a0dc0173a41d3cf6b03a9266c0e23688/image.png
+            # The ordering of faces does not match documentation. See
+            # https://gitlab.kitware.com/vtk/vtk/uploads/a0dc0173a41d3cf6b03a9266c0e23688/image.png
             cell.append(coord_to_vertex(1, 0, 1))
             cell.append(coord_to_vertex(1, 1, 1))
             cell.append(coord_to_vertex(0, 1, 1))
@@ -378,7 +381,7 @@ def test_tetrahedron_mesh_vtk(order):
     check_cell_volume(points, cell, domain, 1 / 6)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', [1, 2, 3, 4])
 def test_quadrilateral_mesh_vtk(order):
     random.seed(13)
@@ -392,7 +395,8 @@ def test_quadrilateral_mesh_vtk(order):
     def coord_to_vertex(x, y):
         return (order + 1) * y + x
 
-    # Make the cell, following https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/  # noqa: E501
+    # Make the cell, following
+    # https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/
     cell = [coord_to_vertex(i, j)
             for i, j in [(0, 0), (order, 0), (order, order), (0, order)]]
     if order > 1:
@@ -417,7 +421,7 @@ def test_quadrilateral_mesh_vtk(order):
     check_cell_volume(points, cell, domain, 1)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', [1, 2, 3, 4])
 def test_hexahedron_mesh_vtk(order):
     if order > 2:
@@ -436,7 +440,8 @@ def test_hexahedron_mesh_vtk(order):
     def coord_to_vertex(x, y, z):
         return (order + 1) ** 2 * z + (order + 1) * y + x
 
-    # Make the cell, following https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/  # noqa: E501
+    # Make the cell, following
+    # https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/
     cell = [coord_to_vertex(x, y, z) for x, y, z in [
         (0, 0, 0), (order, 0, 0), (order, order, 0), (0, order, 0),
         (0, 0, order), (order, 0, order), (order, order, order), (0, order, order)]]
@@ -467,9 +472,10 @@ def test_hexahedron_mesh_vtk(order):
         for i in range(1, order):
             cell.append(coord_to_vertex(0, order, i))
 
-        # The ordering of faces does not match documentation.
-        # See https://gitlab.kitware.com/vtk/vtk/uploads/a0dc0173a41d3cf6b03a9266c0e23688/image.png
-        # The edge flip in this like however has been fixed in VTK so we follow the main documentation link for edges
+        # The ordering of faces does not match documentation. See
+        # https://gitlab.kitware.com/vtk/vtk/uploads/a0dc0173a41d3cf6b03a9266c0e23688/image.png
+        # The edge flip in this like however has been fixed in VTK so we
+        # follow the main documentation link for edges
         for j in range(1, order):
             for i in range(1, order):
                 cell.append(coord_to_vertex(0, i, j))
@@ -502,7 +508,7 @@ def test_hexahedron_mesh_vtk(order):
     check_cell_volume(points, cell, domain, 1)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize("vtk,dolfin,cell_type", [
     ([0, 1, 2, 3, 4, 5], [0, 1, 2, 4, 5, 3], CellType.triangle),
     ([0, 1, 2, 3], [0, 1, 3, 2], CellType.quadrilateral),
@@ -518,15 +524,15 @@ def test_map_vtk_to_dolfin(vtk, dolfin, cell_type):
     assert (cell_p == vtk).all()
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 def test_xdmf_input_tri(datadir):
-    with XDMFFile(MPI.COMM_WORLD, os.path.join(datadir, "mesh.xdmf"), "r", encoding=XDMFFile.Encoding.ASCII) as xdmf:
+    with XDMFFile(MPI.COMM_WORLD, Path(datadir, "mesh.xdmf"), "r", encoding=XDMFFile.Encoding.ASCII) as xdmf:
         mesh = xdmf.read_mesh(name="Grid")
-    surface = assemble_scalar(1 * dx(mesh))
-    assert mesh.mpi_comm().allreduce(surface, op=MPI.SUM) == pytest.approx(4 * np.pi, rel=1e-4)
+    surface = assemble_scalar(form(1 * dx(mesh)))
+    assert mesh.comm.allreduce(surface, op=MPI.SUM) == pytest.approx(4 * np.pi, rel=1e-4)
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', range(1, 4))
 @pytest.mark.parametrize('cell_type', [CellType.triangle, CellType.quadrilateral])
 def test_gmsh_input_2d(order, cell_type):
@@ -569,18 +575,18 @@ def test_gmsh_input_2d(order, cell_type):
 
     cells = cells[:, perm_gmsh(cell_type, cells.shape[1])]
     mesh = create_mesh(MPI.COMM_WORLD, cells, x, ufl_mesh_from_gmsh(gmsh_cell_id, x.shape[1]))
-    surface = assemble_scalar(1 * dx(mesh))
+    surface = assemble_scalar(form(1 * dx(mesh)))
 
-    assert mesh.mpi_comm().allreduce(surface, op=MPI.SUM) == pytest.approx(4 * np.pi, rel=10 ** (-1 - order))
+    assert mesh.comm.allreduce(surface, op=MPI.SUM) == pytest.approx(4 * np.pi, rel=10 ** (-1 - order))
 
     # Bug related to VTK output writing
     # def e2(x):
     #     values = np.empty((x.shape[0], 1))
     #     values[:, 0] = x[:, 0]
     #     return values
-    # cmap = fem.create_coordinate_map(mesh.mpi_comm(), mesh.ufl_domain())
+    # cmap = fem.create_coordinate_map(mesh.comm, mesh.ufl_domain())
     # mesh.geometry.coord_mapping = cmap
-    # V = FunctionSpace(mesh, ("CG", order))
+    # V = FunctionSpace(mesh, ("Lagrange", order))
     # u = Function(V)
     # u.interpolate(e2)
     # from dolfinx.io import VTKFile
@@ -589,7 +595,7 @@ def test_gmsh_input_2d(order, cell_type):
     # print(assemble_scalar(u*dx(mesh)))
 
 
-@skip_in_parallel
+@pytest.mark.skip_in_parallel
 @pytest.mark.parametrize('order', range(1, 4))
 @pytest.mark.parametrize('cell_type', [CellType.tetrahedron, CellType.hexahedron])
 def test_gmsh_input_3d(order, cell_type):
@@ -644,6 +650,26 @@ def test_gmsh_input_3d(order, cell_type):
 
     mesh = create_mesh(MPI.COMM_WORLD, cells, x, domain)
 
-    volume = assemble_scalar(1 * dx(mesh))
+    volume = assemble_scalar(form(1 * dx(mesh)))
 
-    assert mesh.mpi_comm().allreduce(volume, op=MPI.SUM) == pytest.approx(np.pi, rel=10 ** (-1 - order))
+    assert mesh.comm.allreduce(volume, op=MPI.SUM) == pytest.approx(np.pi, rel=10 ** (-1 - order))
+
+
+@pytest.mark.skip_in_parallel
+def test_quadrilateral_cell_order_3():
+    points = [
+        [0., 0.], [1., 0.], [0., 1.], [1., 1.],
+        [1 / 3, 2 / 9], [2 / 3, 2 / 9],
+        [0., 1 / 3], [0., 2 / 3],
+        [1., 1 / 3], [1., 2 / 3],
+        [1 / 3, 1.], [2 / 3, 1.],
+        [1 / 3, 13 / 27], [2 / 3, 13 / 27],
+        [1 / 3, 20 / 27], [2 / 3, 20 / 27]
+    ]
+
+    cell = list(range(16))
+
+    domain = ufl.Mesh(ufl.VectorElement(
+        "Lagrange", ufl.Cell("quadrilateral", geometric_dimension=2), 3))
+
+    check_cell_volume(points, cell, domain, 5 / 6)
